# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Classes for importing AIAA Array Benchmarks.

These classes allow importing data from HDF5 files following the specifications of
the AIAA microphone array methods benchmarking effort:
https://www-docs.b-tu.de/fg-akustik/public/veroeffentlichungen/ArrayMethodsFileFormatsR2P4Release.pdf .

The classes are derived from according Acoular classes so that they can be used directly within
the framework.

Examples
--------
>>> micgeom = MicAIAABenchmark(file='some_benchmarkdata.h5')  # doctest: +SKIP
>>> timedata = TimeSamplesAIAABenchmark(file='some_benchmarkdata.h5')  # doctest: +SKIP


.. autosummary::
    :toctree: generated/

    TimeSamplesAIAABenchmark
    TriggerAIAABenchmark
    CsmAIAABenchmark
    MicAIAABenchmark
"""  # noqa: W505

import contextlib

from numpy import array
from traits.api import (
    File,
    Instance,
    Property,
    cached_property,
    on_trait_change,
    property_depends_on,
)

from acoular.deprecation import deprecated_alias
from acoular.h5files import H5FileBase, _get_h5file_class
from acoular.internal import digest
from acoular.microphones import MicGeom
from acoular.sources import TimeSamples
from acoular.spectra import PowerSpectraImport
from acoular.tools.utils import get_file_basename


class TimeSamplesAIAABenchmark(TimeSamples):
    """Container for AIAA benchmark data in `*.h5` format.

    This class loads measured data from h5 files in AIAA benchmark format
    and and provides information about this data.
    Objects of this class behave similar to :class:`~acoular.sources.TimeSamples`
    objects.
    """

    def _load_timedata(self):
        """Loads timedata from .h5 file. Only for internal use."""
        self.data = self.h5f.get_data_by_reference('MicrophoneData/microphoneDataPa')
        self.sample_freq = self.h5f.get_node_attribute(self.data, 'sampleRateHz')
        (self.num_samples, self.num_channels) = self.data.shape

    def _load_metadata(self):
        """Loads metadata from .h5 file. Only for internal use."""
        self.metadata = {}
        if '/MetaData' in self.h5f:
            self.metadata = self.h5f.node_to_dict('/MetaData')


class TriggerAIAABenchmark(TimeSamplesAIAABenchmark):
    """Container for tacho data in  `*.h5` format.

    This class loads tacho data from h5 files as specified in
    "Microphone Array Benchmark b11: Rotating Point Sources"
    (https://doi.org/10.14279/depositonce-8460)
    and and provides information about this data.
    """

    def _load_timedata(self):
        """Loads timedata from .h5 file. Only for internal use."""
        self.data = self.h5f.get_data_by_reference('TachoData/tachoDataV')
        self.sample_freq = self.h5f.get_node_attribute(self.data, 'sampleRateHz')
        (self.num_samples, self.num_channels) = self.data.shape


@deprecated_alias({'name': 'file'})
class CsmAIAABenchmark(PowerSpectraImport):
    """Class to load the CSM that is stored in AIAA Benchmark HDF5 file."""

    #: Full name of the .h5 file with data
    file = File(filter=['*.h5'], exists=True, desc='name of data file')

    #: Basename of the .h5 file with data, is set automatically.
    basename = Property(
        depends_on=['file'],
        desc='basename of data file',
    )

    #: number of channels
    num_channels = Property()

    #: HDF5 file object
    h5f = Instance(H5FileBase, transient=True)

    # internal identifier
    digest = Property(depends_on=['basename', '_csmsum'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_basename(self):
        return get_file_basename(self.file)

    @on_trait_change('basename')
    def load_data(self):
        """Open the .h5 file and set attributes."""
        if self.h5f is not None:
            with contextlib.suppress(OSError):
                self.h5f.close()
        file = _get_h5file_class()
        self.h5f = file(self.file)

    # @property_depends_on( 'block_size, ind_low, ind_high' )
    def _get_indices(self):
        try:
            return range(self.fftfreq().shape[0])  # [ self.ind_low: self.ind_high ]
        except IndexError:
            return range(0)

    @property_depends_on(['digest'])
    def _get_num_channels(self):
        try:
            attrs = self.h5f.get_data_by_reference('MetaData/ArrayAttributes')
            return self.h5f.get_node_attribute(attrs, 'microphoneCount')
        except IndexError:
            return 0

    @property_depends_on(['digest'])
    def _get_csm(self):
        """Loads cross spectral matrix from file."""
        csmre = self.h5f.get_data_by_reference('/CsmData/csmReal')[:].transpose((2, 0, 1))
        csmim = self.h5f.get_data_by_reference('/CsmData/csmImaginary')[:].transpose((2, 0, 1))
        csmdatagroup = self.h5f.get_data_by_reference('/CsmData')
        sign = self.h5f.get_node_attribute(csmdatagroup, 'fftSign')
        return csmre + sign * 1j * csmim

    def fftfreq(self):
        """Return the Discrete Fourier Transform sample frequencies.

        Returns
        -------
        ndarray
            Array of length *block_size/2+1* containing the sample frequencies.
        """
        return array(self.h5f.get_data_by_reference('/CsmData/binCenterFrequenciesHz')[:].flatten(), dtype=float)


class MicAIAABenchmark(MicGeom):
    """Provides the geometric arrangement of microphones in the array.

    In contrast to standard Acoular microphone geometries, the AIAA
    benchmark format includes the array geometry as metadata in the
    file containing the measurement data.
    """

    #: Name of the .h5-file from which to read the data.
    file = File(filter=['*.h5'], exists=True, desc='name of the h5 file containing the microphone geometry')

    @on_trait_change('file')
    def import_mpos(self):
        """Import the microphone positions from .h5 file.
        Called when :attr:`basename` changes.
        """
        file = _get_h5file_class()
        h5f = file(self.file, mode='r')
        self.pos_total = h5f.get_data_by_reference('MetaData/ArrayAttributes/microphonePositionsM')[:].swapaxes(0, 1)
        h5f.close()



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Contains NUMBA accelerated functions for time-domain beamformers."""

import numba as nb
import numpy as np


@nb.njit(
    [
        (
            nb.float64[:, ::1],
            nb.int64[:, ::1],
            nb.float64[:, ::1],
            nb.float64[:, ::1],
            nb.float64[:, ::1],
            nb.float64[:, ::1],
        ),
    ],
    cache=True,
    parallel=True,
    fastmath=True,
)
def _delayandsum4(data, offsets, ifactor2, steeramp, out, autopower):
    """Performs one time step of delay and sum with output and additional autopower removal.

    Parameters
    ----------
    data : float64[nSamples, nMics]
        The time history for all channels.
    offsets : int64[gridSize, nMics]
        Indices for each grid point and each channel.
    ifactor2: float64[gridSize, nMics]
        Second interpolation factor, the first one is computed internally.
    steeramp: float64[gridSize, nMics]
        Amplitude factor from steering vector.

    Returns
    -------
    None : as the inputs out and autopower get overwritten.

    """
    gridsize, num_channels = offsets.shape
    num = out.shape[0]
    zero_constant = data.dtype.type(0.0)
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            out[n, gi] = zero_constant
            autopower[n, gi] = zero_constant
            for mi in range(num_channels):
                ind = (gi, mi)
                r = (
                    data[offsets[ind] + n, mi] * (1.0 - ifactor2[ind]) + data[offsets[ind] + n + 1, mi] * ifactor2[ind]
                ) * steeramp[ind]
                out[n, gi] += r
                autopower[n, gi] += r * r


@nb.njit(
    [
        (
            nb.float32[:, ::1],
            nb.int32[:, :, ::1],
            nb.float32[:, :, ::1],
            nb.float32[:, :, ::1],
            nb.float32[:, ::1],
            nb.float32[:, ::1],
        ),
        (
            nb.float64[:, ::1],
            nb.int64[:, :, ::1],
            nb.float64[:, :, ::1],
            nb.float64[:, :, ::1],
            nb.float64[:, ::1],
            nb.float64[:, ::1],
        ),
    ],
    cache=True,
    parallel=True,
    fastmath=True,
)
def _delayandsum5(data, offsets, ifactor2, steeramp, out, autopower):
    """Performs one time step of delay and sum with output and additional autopower removal.

    Parameters
    ----------
    data : float64[nSamples, nMics]
        The time history for all channels.
    offsets : int64[nBlockSamples, gridSize, nMics]
        Indices for each grid point and each channel.
    ifactor2: float64[nBlockSamples,gridSize, nMics]
        Second interpolation factor, the first one is computed internally.
    steeramp: float64[nBlockSamples,gridSize, nMics]
        Amplitude factor from steering vector.

    Returns
    -------
    None : as the inputs out and autopower get overwritten.

    """
    num, gridsize, num_channels = offsets.shape
    num = out.shape[0]
    # ZERO = data.dtype.type(0.)
    one_constant = data.dtype.type(1.0)
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            out[n, gi] = 0
            autopower[n, gi] = 0
            for mi in range(num_channels):
                ind = offsets[n, gi, mi] + n
                r = (
                    data[ind, mi] * (one_constant - ifactor2[n, gi, mi]) + data[ind + 1, mi] * ifactor2[n, gi, mi]
                ) * steeramp[
                    n,
                    gi,
                    mi,
                ]
                out[n, gi] += r
                autopower[n, gi] += r * r


@nb.njit(
    [
        (nb.float32[:, :, :], nb.float32[:, :], nb.float32[:, :, :]),
        (nb.float64[:, :, :], nb.float64[:, :], nb.float64[:, :, :]),
    ],
    cache=True,
    parallel=True,
    fastmath=True,
)
def _steer_I(rm, r0, amp):  # noqa: ARG001, N802
    num, gridsize, num_channels = rm.shape
    amp[0, 0, 0] = 1.0 / num_channels  # to get the same type for rm2 as for rm
    nr = amp[0, 0, 0]
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            for mi in nb.prange(num_channels):
                amp[n, gi, mi] = nr


@nb.njit(
    [
        (nb.float32[:, :, :], nb.float32[:, :], nb.float32[:, :, :]),
        (nb.float64[:, :, :], nb.float64[:, :], nb.float64[:, :, :]),
    ],
    cache=True,
    parallel=True,
    fastmath=True,
)
def _steer_II(rm, r0, amp):  # noqa: N802
    num, gridsize, num_channels = rm.shape
    amp[0, 0, 0] = 1.0 / num_channels  # to get the same type for rm2 as for rm
    nr = amp[0, 0, 0]
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            rm2 = np.divide(nr, r0[n, gi])
            for mi in nb.prange(num_channels):
                amp[n, gi, mi] = rm[n, gi, mi] * rm2


@nb.njit(
    [
        (nb.float32[:, :, :], nb.float32[:, :], nb.float32[:, :, :]),
        (nb.float64[:, :, :], nb.float64[:, :], nb.float64[:, :, :]),
    ],
    cache=True,
    parallel=True,
    fastmath=True,
)
def _steer_III(rm, r0, amp):  # noqa: N802
    num, gridsize, num_channels = rm.shape
    rm20 = rm[0, 0, 0] - rm[0, 0, 0]  # to get the same type for rm2 as for rm
    rm1 = rm[0, 0, 0] / rm[0, 0, 0]
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            rm2 = rm20
            for mi in nb.prange(num_channels):
                rm2 += np.divide(rm1, np.square(rm[n, gi, mi]))
            rm2 *= r0[n, gi]
            for mi in nb.prange(num_channels):
                amp[n, gi, mi] = np.divide(rm1, rm[n, gi, mi] * rm2)


@nb.njit(
    [
        (nb.float32[:, :, :], nb.float32[:, :], nb.float32[:, :, :]),
        (nb.float64[:, :, :], nb.float64[:, :], nb.float64[:, :, :]),
    ],
    cache=True,
    parallel=True,
    fastmath=True,
)
def _steer_IV(rm, r0, amp):  # noqa: ARG001, N802
    num, gridsize, num_channels = rm.shape
    amp[0, 0, 0] = np.sqrt(1.0 / num_channels)  # to get the same type for rm2 as for rm
    nr = amp[0, 0, 0]
    rm1 = rm[0, 0, 0] / rm[0, 0, 0]
    rm20 = rm[0, 0, 0] - rm[0, 0, 0]  # to get the same type for rm2 as for rm
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            rm2 = rm20
            for mi in nb.prange(num_channels):
                rm2 += np.divide(rm1, np.square(rm[n, gi, mi]))
            rm2 = np.sqrt(rm2)
            for mi in nb.prange(num_channels):
                amp[n, gi, mi] = np.divide(nr, rm[n, gi, mi] * rm2)


@nb.njit(
    [
        (nb.float32[:, :, ::1], nb.float32, nb.float32[:, :, ::1], nb.int32[:, :, ::1]),
        (nb.float64[:, :, ::1], nb.float64, nb.float64[:, :, ::1], nb.int64[:, :, ::1]),
    ],
    cache=True,
    parallel=True,
    fastmath=True,
)
def _delays(rm, c, interp2, index):
    num, gridsize, num_channels = rm.shape
    invc = 1 / c
    intt = index.dtype.type
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            for mi in nb.prange(num_channels):
                delays = invc * rm[n, gi, mi]
                index[n, gi, mi] = intt(delays)
                interp2[n, gi, mi] = delays - nb.int64(delays)


@nb.njit(
    [
        (nb.float32[:, :, :], nb.float32[:, :, :], nb.int32[:, :, :]),
        (nb.float64[:, :, :], nb.float64[:, :, :], nb.int64[:, :, :]),
    ],
    cache=True,
    parallel=True,
    fastmath=True,
)
def _modf(delays, interp2, index):
    num, gridsize, num_channels = delays.shape
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            for mi in nb.prange(num_channels):
                index[n, gi, mi] = int(delays[n, gi, mi])
                interp2[n, gi, mi] = delays[n, gi, mi] - index[n, gi, mi]


if __name__ == '__main__':
    foo = _delays
    print(foo.parallel_diagnostics(level=4))



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements support separate traits_view definitions for all relevant
classes to lift the traitsui requirement for the Acoular package.
"""

# imports from other packages
from traitsui.api import Item, View
from traitsui.menu import OKCancelButtons

from .base import TimeOut
from .calib import Calib
from .environments import GeneralFlowEnvironment, OpenJet, RotatingFlow, SlotJet, UniformFlowEnvironment
from .fbeamform import (
    BeamformerBase,
    BeamformerCapon,
    BeamformerClean,
    BeamformerCleansc,
    BeamformerCMF,
    BeamformerDamas,
    BeamformerDamasPlus,
    BeamformerEig,
    BeamformerFunctional,
    BeamformerGIB,
    BeamformerMusic,
    BeamformerOrth,
)
from .grids import RectGrid, RectGrid3D
from .microphones import MicGeom
from .process import Average
from .sources import MaskedTimeSamples, SourceMixer, TimeSamples
from .spectra import PowerSpectra
from .tbeamform import BeamformerTime, BeamformerTimeSq, BeamformerTimeSqTraj, BeamformerTimeTraj, IntegratorSectorTime
from .tprocess import FiltFiltOctave, WriteH5, WriteWAV
from .trajectory import Trajectory

MicGeom.class_trait_view(
    'traits_view',
    View(
        [
            'file',
            'num_mics~',
            '|[Microphone geometry]',
        ],
        buttons=OKCancelButtons,
    ),
)

PowerSpectra.class_trait_view(
    'traits_view',
    View(
        [
            'time_data@{}',
            'calib@{}',
            [
                'block_size',
                'window',
                'overlap',
                ['ind_low{Low Index}', 'ind_high{High Index}', '-[Frequency range indices]'],
                ['num_blocks~{Number of blocks}', 'freq_range~{Frequency range}', '-'],
                '[FFT-parameters]',
            ],
        ],
        buttons=OKCancelButtons,
    ),
)

Calib.class_trait_view(
    'traits_view',
    View(
        [
            'file{File name}',
            [
                'num_mics~{Number of microphones}',
                '|[Properties]',
            ],
        ],
        title='Calibration data',
        buttons=OKCancelButtons,
    ),
)

Trajectory.class_trait_view(
    'traits_view',
    View(
        [
            Item('points', style='custom'),
        ],
        title='Grid center trajectory',
        buttons=OKCancelButtons,
    ),
)

RectGrid.class_trait_view(
    'traits_view',
    View(
        [
            ['x_min', 'y_min', '|'],
            ['x_max', 'y_max', 'z', 'increment', 'size~{Grid size}', '|'],
            '-[Map extension]',
        ],
    ),
)

# increment3D omitted in view for easier handling, can be added later
RectGrid3D.class_trait_view(
    'traits_view',
    View(
        [
            ['x_min', 'y_min', 'z_min', '|'],
            ['x_max', 'y_max', 'z_max', 'increment', 'size~{Grid size}', '|'],
            '-[Map extension]',
        ],
    ),
)

BeamformerTime.class_trait_view(
    'traits_view',
    View(
        [
            [Item('steer{}', style='custom')],
            [Item('source{}', style='custom'), '-<>'],
            [Item('weights{}', style='simple')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerTimeSq.class_trait_view(
    'traits_view',
    View(
        [
            [Item('steer{}', style='custom')],
            [Item('source{}', style='custom'), '-<>'],
            [Item('r_diag', label='diagonal removed')],
            [Item('weights{}', style='simple')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerTimeTraj.class_trait_view(
    'traits_view',
    View(
        [
            [Item('steer{}', style='custom')],
            [Item('source{}', style='custom'), '-<>'],
            [Item('trajectory{}', style='custom')],
            [Item('weights{}', style='simple')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerTimeSqTraj.class_trait_view(
    'traits_view',
    View(
        [
            [Item('steer{}', style='custom')],
            [Item('source{}', style='custom'), '-<>'],
            [Item('trajectory{}', style='custom')],
            [Item('r_diag', label='diagonal removed')],
            [Item('weights{}', style='simple')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

IntegratorSectorTime.class_trait_view(
    'traits_view',
    View(
        [
            [Item('sectors', style='custom')],
            [Item('grid', style='custom'), '-<>'],
            '|',
        ],
        title='Integrator',
        buttons=OKCancelButtons,
    ),
)

UniformFlowEnvironment.class_trait_view(
    'traits_view',
    View(
        [
            ['ma{Flow Mach number}', 'fdv{Flow vector}'],
            '|[Uniform Flow]',
        ],
    ),
)

GeneralFlowEnvironment.class_trait_view(
    'traits_view',
    View(
        [
            ['ff{Flow field}', 'N{Max. number of rays}', 'Om{Max. solid angle }'],
            '|[General Flow]',
        ],
    ),
)

SlotJet.class_trait_view(
    'traits_view',
    View(
        [
            ['v0{Exit velocity}', 'origin{Jet origin}', 'flow', 'plane', 'B{Slot width}'],
            '|[Slot jet]',
        ],
    ),
)

OpenJet.class_trait_view(
    'traits_view',
    View(
        [
            ['v0{Exit velocity}', 'origin{Jet origin}', 'D{Nozzle diameter}'],
            '|[Open jet]',
        ],
    ),
)

RotatingFlow.class_trait_view(
    'traits_view',
    View(
        [
            ['v0{flow velocity}', 'origin{Jet origin}', 'rpm{ revolutions }'],
            '|[RotatingFlow]',
        ],
    ),
)

TimeOut.class_trait_view(
    'traits_view',
    View(
        Item('source', style='custom'),
    ),
)

Average.class_trait_view(
    'traits_view',
    View(
        [
            Item('source', style='custom'),
            'naverage{Samples to average}',
            ['sample_freq~{Output sampling frequency}', '|[Properties]'],
            '|',
        ],
        title='Linear average',
        buttons=OKCancelButtons,
    ),
)

FiltFiltOctave.class_trait_view(
    'traits_view',
    View(
        [
            Item('source', style='custom'),
            'band{Center frequency}',
            'fraction{Bandwidth}',
            ['sample_freq~{Output sampling frequency}', '|[Properties]'],
            '|',
        ],
        title='Linear average',
        buttons=OKCancelButtons,
    ),
)

WriteWAV.class_trait_view(
    'traits_view',
    View(
        [
            Item('source', style='custom'),
            ['basename~{File name}', '|[Properties]'],
            '|',
        ],
        title='Write wav file',
        buttons=OKCancelButtons,
    ),
)

WriteH5.class_trait_view(
    'traits_view',
    View(
        [
            Item('source', style='custom'),
            ['name{File name}', '|[Properties]'],
            '|',
        ],
        title='write .h5',
        buttons=OKCancelButtons,
    ),
)

TimeSamples.class_trait_view(
    'traits_view',
    View(
        [
            'name{File name}',
            [
                'sample_freq~{Sampling frequency}',
                'num_channels~{Number of channels}',
                'num_samples~{Number of samples}',
                '|[Properties]',
            ],
            '|',
        ],
        title='Time data',
        buttons=OKCancelButtons,
    ),
)

MaskedTimeSamples.class_trait_view(
    'traits_view',
    View(
        [
            'name{File name}',
            ['start{From sample}', Item('stop', label='to', style='text'), '-'],
            'invalid_channels{Invalid channels}',
            [
                'sample_freq~{Sampling frequency}',
                'num_channels~{Number of channels}',
                'num_samples~{Number of samples}',
                '|[Properties]',
            ],
            '|',
        ],
        title='Time data',
        buttons=OKCancelButtons,
    ),
)

SourceMixer.class_trait_view(
    'traits_view',
    View(
        Item('sources', style='custom'),
    ),
)


BeamformerBase.class_trait_view(
    'traits_view',
    View(
        [
            [Item('r_diag', label='Diagonal removed')],
            [Item('steer', label='Steering vector')],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerFunctional.class_trait_view(
    'traits_view',
    View(
        [
            #            [Item('mics{}', style='custom')],
            #            [Item('grid', style='custom'), '-<>'],
            [Item('gamma', label='Exponent', style='simple')],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerCapon.class_trait_view(
    'traits_view',
    View(
        [
            #            [Item('mics{}', style='custom')],
            #            [Item('grid', style='custom'), '-<>'],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerEig.class_trait_view(
    'traits_view',
    View(
        [
            #            [Item('mics{}', style='custom')],
            #            [Item('grid', style='custom'), '-<>'],
            [Item('n', label='Component No.', style='simple')],
            [Item('r_diag', label='Diagonal removed')],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerMusic.class_trait_view(
    'traits_view',
    View(
        [
            #            [Item('mics{}', style='custom')],
            #            [Item('grid', style='custom'), '-<>'],
            [Item('n', label='No. of sources', style='simple')],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerDamas.class_trait_view(
    'traits_view',
    View(
        [
            [Item('beamformer{}', style='custom')],
            [Item('n_iter{Number of iterations}')],
            #            [Item('steer{Type of steering vector}')],
            [Item('calcmode{How to calculate PSF}')],
            '|',
        ],
        title='Beamformer denconvolution options',
        buttons=OKCancelButtons,
    ),
)

BeamformerDamasPlus.class_trait_view(
    'traits_view',
    View(
        [
            [Item('beamformer{}', style='custom')],
            [Item('method{Solver}')],
            [Item('n_iter{Max. number of iterations}')],
            [Item('alpha', label='Lasso weight factor')],
            [Item('calcmode{How to calculate PSF}')],
            '|',
        ],
        title='Beamformer denconvolution options',
        buttons=OKCancelButtons,
    ),
)

BeamformerOrth.class_trait_view(
    'traits_view',
    View(
        [
            #            [Item('mpos{}', style='custom')],
            #            [Item('grid', style='custom'), '-<>'],
            [Item('n', label='Number of components', style='simple')],
            [Item('r_diag', label='Diagonal removed')],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerCleansc.class_trait_view(
    'traits_view',
    View(
        [
            #            [Item('mpos{}', style='custom')],
            #            [Item('grid', style='custom'), '-<>'],
            [Item('n_iter', label='No. of iterations', style='simple')],
            [Item('r_diag', label='Diagonal removed')],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerClean.class_trait_view(
    'traits_view',
    View(
        [
            [Item('beamformer{}', style='custom')],
            [Item('n_iter{Number of iterations}')],
            #            [Item('steer{Type of steering vector}')],
            [Item('calcmode{How to calculate PSF}')],
            '|',
        ],
        title='Beamformer denconvolution options',
        buttons=OKCancelButtons,
    ),
)

BeamformerCMF.class_trait_view(
    'traits_view',
    View(
        [
            #            [Item('mpos{}', style='custom')],
            #            [Item('grid', style='custom'), '-<>'],
            [Item('method', label='Fit method')],
            [Item('n_iter', label='(Max.) no. of iterations')],
            [Item('alpha', label='Lasso weight factor')],
            [Item('c', label='Speed of sound')],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)

BeamformerGIB.class_trait_view(
    'traits_view',
    View(
        [
            #            [Item('mpos{}', style='custom')],
            #            [Item('grid', style='custom'), '-<>'],
            [Item('method', label='Fit method')],
            [Item('n_iter', label='(Max.) no. of iterations')],
            [Item('alpha', label='Lasso weight factor')],
            [Item('c', label='Speed of sound')],
            #            [Item('env{}', style='custom')],
            '|',
        ],
        title='Beamformer options',
        buttons=OKCancelButtons,
    ),
)



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------

from .configuration import config


class H5FileBase:
    """Base class for File objects that handle writing and reading of .h5 files."""

    def create_extendable_array(self, nodename, shape, precision, group=None):
        pass

    def get_data_by_reference(self, nodename, group=None):
        pass

    def set_node_attribute(self, node, attrname, value):
        pass

    def get_node_attribute(self, node, attrname):
        pass

    def append_data(self, node, data):
        pass

    def remove_data(self, nodename):
        pass

    def create_new_group(self, name, group=None):
        pass


class H5CacheFileBase:
    """Base class for File objects that handle writing and reading of .h5 cache files."""

    compression_filter = None

    def is_cached(self, nodename, group=None):
        pass

    def create_compressible_array(self, nodename, shape, precision, group=None):
        pass


if config.have_tables:
    import tables

    precision_to_atom = {
        'float32': tables.Float32Atom(),
        'complex64': tables.ComplexAtom(8),
        'float64': tables.Float64Atom(),
        'complex128': tables.ComplexAtom(16),
        'bool': tables.BoolAtom(),
        'int32': tables.Int32Atom(),
        'int16': tables.Int16Atom(),
        'int8': tables.Int8Atom(),
    }

    class H5FileTables(H5FileBase, tables.File):
        def create_extendable_array(self, nodename, shape, precision, group=None):
            if not group:
                group = self.root
            atom = precision_to_atom[precision]
            self.create_earray(group, nodename, atom, shape)

        def get_data_by_reference(self, nodename, group=None):
            if not group:
                group = self.root
            return self.get_node(group, nodename)

        def set_node_attribute(self, node, attrname, value):
            node.set_attr(attrname, value)

        def get_node_attribute(self, node, attrname):
            return node._v_attrs[attrname]  # noqa: SLF001

        def append_data(self, node, data):
            node.append(data)

        def remove_data(self, nodename):
            self.remove_node('/', nodename, recursive=True)

        def create_new_group(self, name, group=None):
            if not group:
                group = self.root
            return self.create_group(group, name)

        def get_child_nodes(self, nodename):
            for childnode in self.list_nodes(nodename):
                yield (childnode.name, childnode)

        def node_to_dict(self, nodename):
            """Recursively convert an HDF5 node to a dictionary."""
            node = self.get_node(nodename)
            # initialize node-dict with node's own attributes
            result = {attr: node._v_attrs[attr] for attr in node._v_attrs._f_list()}  # noqa: SLF001
            if isinstance(node, tables.Group):
                # if node is a group, recursively add its children
                for childname in node._v_children:  # noqa: SLF001
                    result[childname] = self.node_to_dict(f'{nodename}/{childname}')
            elif isinstance(node, tables.Leaf):
                # if node contains only data, add it
                return node
            else:
                return None
            return result

    class H5CacheFileTables(H5FileTables, H5CacheFileBase):
        compression_filter = tables.Filters(complevel=5, complib='blosc')

        def is_cached(self, nodename, group=None):
            if not group:
                group = self.root
            return nodename in group

        def create_compressible_array(self, nodename, shape, precision, group=None):
            if not group:
                group = self.root
            atom = precision_to_atom[precision]
            self.create_carray(group, nodename, atom, shape, filters=self.compression_filter)


if config.have_h5py:
    import h5py

    class H5FileH5py(H5FileBase, h5py.File):
        def _get_in_file_path(self, nodename, group=None):
            if not group:
                return '/' + nodename
            return group + '/' + nodename

        def create_array(self, where, name, obj):
            self.create_dataset(f'{where}/{name}', data=obj)

        def create_extendable_array(self, nodename, shape, precision, group=None):
            in_file_path = self._get_in_file_path(nodename, group)
            self.create_dataset(in_file_path, shape=shape, dtype=precision, maxshape=(None, shape[1]))

        def get_data_by_reference(self, nodename, group=None):
            in_file_path = self._get_in_file_path(nodename, group)
            return self[in_file_path]

        def set_node_attribute(self, node, attrname, value):
            node.attrs[attrname] = value

        def get_node_attribute(self, node, attrname):
            return node.attrs[attrname]

        def append_data(self, node, data):
            old_shape = node.shape
            new_shape = (old_shape[0] + data.shape[0], data.shape[1])
            node.resize(new_shape)
            node[old_shape[0] : new_shape[0], :] = data

        def remove_data(self, nodename, group=None):
            in_file_path = self._get_in_file_path(nodename, group)
            del self[in_file_path]

        def create_new_group(self, name, group=None):
            in_file_path = self._get_in_file_path(name, group)
            self.create_group(in_file_path)
            return in_file_path

        def get_child_nodes(self, nodename):
            for childnode in self[nodename]:
                yield (childnode, self[f'{nodename}/{childnode}'])

        def node_to_dict(self, nodename):
            """Recursively convert an HDF5 node to a dictionary."""
            node = self[nodename]
            # initialize node-dict with node's own attributes
            result = {attr: node.attrs[attr] for attr in node.attrs}
            if isinstance(node, h5py.Group):
                # if node is a group, recursively add its children
                for childname in node:
                    result[childname] = self.node_to_dict(f'{nodename}/{childname}')
            elif isinstance(node, h5py.Dataset):
                # if node contains only data, add it
                return node
            else:
                return None
            return result

    class H5CacheFileH5py(H5CacheFileBase, H5FileH5py):
        compression_filter = 'lzf'
        #        compression_filter = 'blosc' # unavailable...

        def is_cached(self, nodename, group=None):
            if not group:
                group = '/'
            return group + nodename in self

        def create_compressible_array(self, nodename, shape, precision, group=None):
            in_file_path = self._get_in_file_path(nodename, group)
            self.create_dataset(
                in_file_path,
                dtype=precision,
                shape=shape,
                compression=self.compression_filter,
                chunks=True,
            )


def _get_h5file_class():
    if config.h5library in ['pytables', 'tables']:
        return H5FileTables
    if config.h5library == 'h5py':
        return H5FileH5py
    return None


def _get_cachefile_class():
    if config.h5library in ['pytables', 'tables']:
        return H5CacheFileTables
    if config.h5library == 'h5py':
        return H5CacheFileH5py
    return None



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements blockwise processing in the time domain.

.. autosummary::
    :toctree: generated/

    MaskedTimeOut
    Trigger
    AngleTracker
    ChannelMixer
    SpatialInterpolator
    SpatialInterpolatorRotation
    SpatialInterpolatorConstantRotation
    Mixer
    TimePower
    TimeCumAverage
    TimeReverse
    Filter
    FilterBank
    FiltFiltOctave
    FiltOctave
    TimeExpAverage
    FiltFreqWeight
    OctaveFilterBank
    WriteWAV
    WriteH5
    TimeConvolve
    MaskedTimeInOut
"""

# imports from other packages
import wave
from abc import abstractmethod
from datetime import datetime, timezone
from os import path
from warnings import warn

import numba as nb
from numpy import (
    append,
    arange,
    argmax,
    argmin,
    argsort,
    array,
    array_equal,
    asarray,
    ceil,
    concatenate,
    cumsum,
    delete,
    empty,
    empty_like,
    exp,
    flatnonzero,
    float64,
    identity,
    inf,
    int16,
    interp,
    linspace,
    mean,
    nan,
    newaxis,
    pi,
    polymul,
    sin,
    sinc,
    split,
    sqrt,
    stack,
    sum,  # noqa: A004
    tile,
    unique,
    zeros,
)
from scipy.fft import irfft, rfft
from scipy.interpolate import CloughTocher2DInterpolator, CubicSpline, LinearNDInterpolator, Rbf, splev, splrep
from scipy.linalg import norm
from scipy.signal import bilinear, butter, sosfilt, sosfiltfilt, tf2sos
from scipy.spatial import Delaunay
from traits.api import (
    Bool,
    CArray,
    CInt,
    Constant,
    Delegate,
    Dict,
    Enum,
    File,
    Float,
    Instance,
    Int,
    List,
    Map,
    Property,
    Str,
    Union,
    cached_property,
    observe,
    on_trait_change,
)

# acoular imports
from .base import SamplesGenerator, TimeOut
from .configuration import config
from .deprecation import deprecated_alias
from .environments import cartToCyl, cylToCart
from .h5files import _get_h5file_class
from .internal import digest, ldigest
from .microphones import MicGeom
from .tools.utils import find_basename


@deprecated_alias({'numchannels_total': 'num_channels_total', 'numsamples_total': 'num_samples_total'})
class MaskedTimeOut(TimeOut):
    """Signal processing block for channel and sample selection.

    This class serves as intermediary to define (in)valid
    channels and samples for any
    :class:`~acoular.sources.SamplesGenerator` (or derived) object.
    It gets samples from :attr:`~acoular.base.TimeOut.source`
    and generates output via the generator :meth:`result`.
    """

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Index of the first sample to be considered valid.
    start = CInt(0, desc='start of valid samples')

    # Index of the last sample to be considered valid.
    stop = Union(None, CInt, desc='stop of valid samples')

    # Channels that are to be treated as invalid.
    invalid_channels = List(int, desc='list of invalid channels')

    # Channel mask to serve as an index for all valid channels, is set automatically.
    channels = Property(depends_on=['invalid_channels', 'source.num_channels'], desc='channel mask')

    # Number of channels in input, as given by :attr:`~acoular.base.TimeOut.source`.
    num_channels_total = Delegate('source', 'num_channels')

    # Number of samples in input, as given by :attr:`~acoular.base.TimeOut.source`.
    num_samples_total = Delegate('source', 'num_samples')

    # Number of valid channels, is set automatically.
    num_channels = Property(
        depends_on=['invalid_channels', 'source.num_channels'], desc='number of valid input channels'
    )

    # Number of valid time samples, is set automatically.
    num_samples = Property(
        depends_on=['start', 'stop', 'source.num_samples'], desc='number of valid samples per channel'
    )

    # Name of the cache file without extension, readonly.
    basename = Property(depends_on=['source.digest'], desc='basename for cache file')

    # internal identifier
    digest = Property(depends_on=['source.digest', 'start', 'stop', 'invalid_channels'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_basename(self):
        warn(
            (
                f'The basename attribute of a {self.__class__.__name__} object is deprecated'
                ' and will be removed in a future release!'
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return find_basename(self.source, alternative_basename=self.source.__class__.__name__ + self.source.digest)

    @cached_property
    def _get_channels(self):
        if len(self.invalid_channels) == 0:
            return slice(0, None, None)
        allr = [i for i in range(self.num_channels_total) if i not in self.invalid_channels]
        return array(allr)

    @cached_property
    def _get_num_channels(self):
        if len(self.invalid_channels) == 0:
            return self.num_channels_total
        return len(self.channels)

    @cached_property
    def _get_num_samples(self):
        sli = slice(self.start, self.stop).indices(self.num_samples_total)
        return sli[1] - sli[0]

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, :attr:`num_channels`).
            The last block may be shorter than num.

        """
        sli = slice(self.start, self.stop).indices(self.num_samples_total)
        start = sli[0]
        stop = sli[1]
        if start >= stop:
            msg = 'no samples available'
            raise OSError(msg)

        if start != 0 or stop != self.num_samples_total:
            offset = -start % num
            if offset == 0:
                offset = num
            buf = empty((num + offset, self.num_channels), dtype=float)
            bsize = 0
            i = 0
            fblock = True
            for block in self.source.result(num):
                bs = block.shape[0]
                i += bs
                if fblock and i >= start:  # first block in the chosen interval
                    if i >= stop:  # special case that start and stop are in one block
                        yield block[bs - (i - start) : bs - (i - stop), self.channels]
                        break
                    bsize += i - start
                    buf[: (i - start), :] = block[bs - (i - start) :, self.channels]
                    fblock = False
                elif i >= stop:  # last block
                    buf[bsize : bsize + bs - (i - stop), :] = block[: bs - (i - stop), self.channels]
                    bsize += bs - (i - stop)
                    if bsize > num:
                        yield buf[:num]
                        buf[: bsize - num, :] = buf[num:bsize, :]
                        bsize -= num
                    yield buf[:bsize, :]
                    break
                elif i >= start:
                    buf[bsize : bsize + bs, :] = block[:, self.channels]
                    bsize += bs
                if bsize >= num:
                    yield buf[:num]
                    buf[: bsize - num, :] = buf[num:bsize, :]
                    bsize -= num

        else:  # if no start/stop given, don't do the resorting thing
            for block in self.source.result(num):
                yield block[:, self.channels]


class ChannelMixer(TimeOut):
    """Class for directly mixing the channels of a multi-channel source.
    Outputs a single channel.
    """

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Amplitude weight(s) for the channels as array. If not set, all channels are equally weighted.
    weights = CArray(desc='channel weights')

    # Number of channels is always one here.
    num_channels = Constant(1)

    # internal identifier
    digest = Property(depends_on=['source.digest', 'weights'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, 1).
            The last block may be shorter than num.

        """
        if self.weights.size:
            if self.weights.shape in {(self.source.num_channels,), (1,)}:
                weights = self.weights
            else:
                msg = f'Weight factors can not be broadcasted: {self.weights.shape}, {(self.source.num_channels,)}'
                raise ValueError(msg)
        else:
            weights = 1

        for block in self.source.result(num):
            yield sum(weights * block, 1, keepdims=True)


class Trigger(TimeOut):  # pragma: no cover
    """Class for identifying trigger signals.
    Gets samples from :attr:`source` and stores the trigger samples in :meth:`trigger_data`.

    The algorithm searches for peaks which are above/below a signed threshold.
    A estimate for approximative length of one revolution is found via the greatest
    number of samples between the adjacent peaks.
    The algorithm then defines hunks as percentages of the estimated length of one revolution.
    If there are multiple peaks within one hunk, the algorithm just takes one of them
    into account (e.g. the first peak, the peak with extremum value, ...).
    In the end, the algorithm checks if the found peak locations result in rpm that don't
    vary too much.
    """

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Threshold of trigger. Has different meanings for different
    # :attr:`~acoular.tprocess.Trigger.trigger_type`. The sign is relevant.
    # If a sample of the signal is above/below the positive/negative threshold,
    # it is assumed to be a peak.
    # Default is None, in which case a first estimate is used: The threshold
    # is assumed to be 75% of the max/min difference between all extremums and the
    # mean value of the trigger signal. E.g: the mean value is 0 and there are positive
    # extremums at 400 and negative extremums at -800. Then the estimated threshold would be
    # 0.75 * -800 = -600.
    threshold = Float(None)

    # Maximum allowable variation of length of each revolution duration. Default is
    # 2%. A warning is thrown, if any revolution length surpasses this value:
    # abs(durationEachRev - meanDuration) > 0.02 * meanDuration
    max_variation_of_duration = Float(0.02)

    # Defines the length of hunks via lenHunk = hunk_length * maxOncePerRevDuration.
    # If there are multiple peaks within lenHunk, then the algorithm will
    # cancel all but one out (see :attr:`~acoular.tprocess.Trigger.multiple_peaks_in_hunk`).
    # Default is to 0.1.
    hunk_length = Float(0.1)

    # Type of trigger.
    #
    # 'dirac': a single pulse is assumed (sign of :attr:`~acoular.tprocess.Trigger.trigger_type` is
    # important). Sample will trigger if its value is above/below the pos/neg threshold.
    #
    # 'rect' : repeating rectangular functions. Only every second edge is assumed to be a trigger.
    # The sign of :attr:`~acoular.tprocess.Trigger.trigger_type` gives information on which edge
    # should be used (+ for rising edge, - for falling edge). Sample will trigger if the difference
    # between its value and its predecessors value is above/below the pos/neg threshold.
    #
    # Default is 'dirac'.
    trigger_type = Enum('dirac', 'rect')

    # Identifier which peak to consider, if there are multiple peaks in one hunk : (see
    # :attr:`~acoular.tprocess.Trigger.hunk_length`). Default is to 'extremum', : in which case the
    # extremal peak (maximum if threshold > 0, minimum if threshold < 0) is considered.
    multiple_peaks_in_hunk = Enum('extremum', 'first')

    # Tuple consisting of 3 entries:
    #
    # 1.: -Vector with the sample indices of the 1/Rev trigger samples
    #
    # 2.: -maximum of number of samples between adjacent trigger samples
    #
    # 3.: -minimum of number of samples between adjacent trigger samples
    trigger_data = Property(
        depends_on=[
            'source.digest',
            'threshold',
            'max_variation_of_duration',
            'hunk_length',
            'trigger_type',
            'multiple_peaks_in_hunk',
        ],
    )

    # internal identifier
    digest = Property(
        depends_on=[
            'source.digest',
            'threshold',
            'max_variation_of_duration',
            'hunk_length',
            'trigger_type',
            'multiple_peaks_in_hunk',
        ],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_trigger_data(self):
        self._check_trigger_existence()
        triggerFunc = {'dirac': self._trigger_dirac, 'rect': self._trigger_rect}[self.trigger_type]
        num = 2048  # number samples for result-method of source
        threshold = self._threshold(num)

        # get all samples which surpasse the threshold
        peakLoc = array([], dtype='int')  # all indices which surpasse the threshold
        trigger_data = array([])
        x0 = []
        dSamples = 0
        for triggerSignal in self.source.result(num):
            localTrigger = flatnonzero(triggerFunc(x0, triggerSignal, threshold))
            if len(localTrigger) != 0:
                peakLoc = append(peakLoc, localTrigger + dSamples)
                trigger_data = append(trigger_data, triggerSignal[localTrigger])
            dSamples += num
            x0 = triggerSignal[-1]
        if len(peakLoc) <= 1:
            msg = 'Not enough trigger info. Check *threshold* sign and value!'
            raise Exception(msg)

        peakDist = peakLoc[1:] - peakLoc[:-1]
        maxPeakDist = max(peakDist)  # approximate distance between the revolutions

        # if there are hunks which contain multiple peaks -> check for each hunk,
        # which peak is the correct one -> delete the other one.
        # if there are no multiple peaks in any hunk left -> leave the while
        # loop and continue with program
        multiplePeaksWithinHunk = flatnonzero(peakDist < self.hunk_length * maxPeakDist)
        while len(multiplePeaksWithinHunk) > 0:
            peakLocHelp = multiplePeaksWithinHunk[0]
            indHelp = [peakLocHelp, peakLocHelp + 1]
            if self.multiple_peaks_in_hunk == 'extremum':
                values = trigger_data[indHelp]
                deleteInd = indHelp[argmin(abs(values))]
            elif self.multiple_peaks_in_hunk == 'first':
                deleteInd = indHelp[1]
            peakLoc = delete(peakLoc, deleteInd)
            trigger_data = delete(trigger_data, deleteInd)
            peakDist = peakLoc[1:] - peakLoc[:-1]
            multiplePeaksWithinHunk = flatnonzero(peakDist < self.hunk_length * maxPeakDist)

        # check whether distances between peaks are evenly distributed
        meanDist = mean(peakDist)
        diffDist = abs(peakDist - meanDist)
        faultyInd = flatnonzero(diffDist > self.max_variation_of_duration * meanDist)
        if faultyInd.size != 0:
            warn(
                f'In Trigger-Identification: The distances between the peaks (and therefore the lengths of the \
                revolutions) vary too much (check samples {peakLoc[faultyInd] + self.source.start}).',
                Warning,
                stacklevel=2,
            )
        return peakLoc, max(peakDist), min(peakDist)

    def _trigger_dirac(self, x0, x, threshold):  # noqa: ARG002
        # x0 not needed here, but needed in _trigger_rect
        return self._trigger_value_comp(x, threshold)

    def _trigger_rect(self, x0, x, threshold):
        # x0 stores the last value of the the last generator cycle
        xNew = append(x0, x)
        # indPeakHunk = abs(xNew[1:] - xNew[:-1]) > abs(threshold)
        # with above line, every edge would be located
        return self._trigger_value_comp(xNew[1:] - xNew[:-1], threshold)

    def _trigger_value_comp(self, trigger_data, threshold):
        return trigger_data > threshold if threshold > 0.0 else trigger_data < threshold

    def _threshold(self, num):
        if self.threshold is None:  # take a guessed threshold
            # get max and min values of whole trigger signal
            maxVal = -inf
            minVal = inf
            meanVal = 0
            cntMean = 0
            for trigger_data in self.source.result(num):
                maxVal = max(maxVal, trigger_data.max())
                minVal = min(minVal, trigger_data.min())
                meanVal += trigger_data.mean()
                cntMean += 1
            meanVal /= cntMean

            # get 75% of maximum absolute value of trigger signal
            maxTriggerHelp = [minVal, maxVal] - meanVal
            argInd = argmax(abs(maxTriggerHelp))
            thresh = maxTriggerHelp[argInd] * 0.75  # 0.75 for 75% of max trigger signal
            warn(f'No threshold was passed. An estimated threshold of {thresh} is assumed.', Warning, stacklevel=2)
        else:  # take user defined  threshold
            thresh = self.threshold
        return thresh

    def _check_trigger_existence(self):
        nChannels = self.source.num_channels
        if nChannels != 1:
            msg = f'Trigger signal must consist of ONE channel, instead {nChannels} channels are given!'
            raise Exception(msg)
        return 0

    def result(self, num):
        msg = 'result method not implemented yet! Data from source will be passed without transformation.'
        warn(msg, Warning, stacklevel=2)
        yield from self.source.result(num)


class AngleTracker(MaskedTimeOut):
    """Calculates rotation angle and rpm per sample from a trigger signal
    using spline interpolation in the time domain.

    Gets samples from :attr:`trigger` and stores the angle and rpm samples in :meth:`angle` and
    :meth:`rpm`.
    """

    # Trigger data from :class:`acoular.tprocess.Trigger`.
    trigger = Instance(Trigger)

    # internal identifier
    digest = Property(
        depends_on=[
            'source.digest',
            'trigger.digest',
            'trigger_per_revo',
            'rot_direction',
            'interp_points',
            'start_angle',
        ],
    )

    # Trigger signals per revolution,
    # defaults to 1.
    trigger_per_revo = Int(1, desc='trigger signals per revolution')

    # Flag to set counter-clockwise (1) or clockwise (-1) rotation,
    # defaults to -1.
    rot_direction = Int(-1, desc='mathematical direction of rotation')

    # Points of interpolation used for spline,
    # defaults to 4.
    interp_points = Int(4, desc='Points of interpolation used for spline')

    # rotation angle in radians for first trigger position
    start_angle = Float(0, desc='rotation angle for trigger position')

    # revolutions per minute for each sample, read-only
    rpm = Property(depends_on=['digest'], desc='revolutions per minute for each sample')

    # average revolutions per minute, read-only
    average_rpm = Property(depends_on=['digest'], desc='average revolutions per minute')

    # rotation angle in radians for each sample, read-only
    angle = Property(depends_on=['digest'], desc='rotation angle for each sample')

    # Internal flag to determine whether rpm and angle calculation has been processed,
    # prevents recalculation
    _calc_flag = Bool(False)

    # Revolutions per minute, internal use
    _rpm = CArray()

    # Rotation angle in radians, internal use
    _angle = CArray()

    @cached_property
    def _get_digest(self):
        return digest(self)

    # helperfunction for trigger index detection
    def _find_nearest_idx(self, peakarray, value):
        peakarray = asarray(peakarray)
        return (abs(peakarray - value)).argmin()

    def _to_rpm_and_angle(self):
        """Internal helper function
        Calculates angles in radians for one or more instants in time.

        Current version supports only trigger and sources with the same samplefreq.
        This behaviour may change in future releases
        """
        # init
        ind = 0
        # trigger data
        peakloc, maxdist, mindist = self.trigger.trigger_data
        TriggerPerRevo = self.trigger_per_revo
        rotDirection = self.rot_direction
        num = self.source.num_samples
        samplerate = self.source.sample_freq
        self._rpm = zeros(num)
        self._angle = zeros(num)
        # number of spline points
        InterpPoints = self.interp_points

        # loop over all timesamples
        while ind < num:
            # when starting spline forward
            if ind < peakloc[InterpPoints]:
                peakdist = (
                    peakloc[self._find_nearest_idx(peakarray=peakloc, value=ind) + 1]
                    - peakloc[self._find_nearest_idx(peakarray=peakloc, value=ind)]
                )
                splineData = stack(
                    (range(InterpPoints), peakloc[ind // peakdist : ind // peakdist + InterpPoints]),
                    axis=0,
                )
            # spline backwards
            else:
                peakdist = (
                    peakloc[self._find_nearest_idx(peakarray=peakloc, value=ind)]
                    - peakloc[self._find_nearest_idx(peakarray=peakloc, value=ind) - 1]
                )
                splineData = stack(
                    (range(InterpPoints), peakloc[ind // peakdist - InterpPoints : ind // peakdist]),
                    axis=0,
                )
            # calc angles and rpm
            Spline = splrep(splineData[:, :][1], splineData[:, :][0], k=3)
            self._rpm[ind] = splev(ind, Spline, der=1, ext=0) * 60 * samplerate
            self._angle[ind] = (
                splev(ind, Spline, der=0, ext=0) * 2 * pi * rotDirection / TriggerPerRevo + self.start_angle
            ) % (2 * pi)
            # next sample
            ind += 1
        # calculation complete
        self._calc_flag = True

    # reset calc flag if something has changed
    @on_trait_change('digest')
    def _reset_calc_flag(self):
        self._calc_flag = False

    # calc rpm from trigger data
    @cached_property
    def _get_rpm(self):
        if not self._calc_flag:
            self._to_rpm_and_angle()
        return self._rpm

    # calc of angle from trigger data
    @cached_property
    def _get_angle(self):
        if not self._calc_flag:
            self._to_rpm_and_angle()
        return self._angle

    # calc average rpm from trigger data
    @cached_property
    def _get_average_rpm(self):
        """Returns average revolutions per minute (rpm) over the source samples.

        Returns
        -------
        rpm : float
            rpm in 1/min.

        """
        # trigger indices data
        peakloc = self.trigger.trigger_data[0]
        # calculation of average rpm in 1/min
        return (len(peakloc) - 1) / (peakloc[-1] - peakloc[0]) / self.trigger_per_revo * self.source.sample_freq * 60


class SpatialInterpolator(TimeOut):  # pragma: no cover
    """Base class for spatial interpolation of microphone data.
    Gets samples from :attr:`source` and generates output via the
    generator :meth:`result`.
    """

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # :class:`~acoular.microphones.MicGeom` object that provides the real microphone locations.
    mics = Instance(MicGeom(), desc='microphone geometry')

    # :class:`~acoular.microphones.MicGeom` object that provides the virtual microphone locations.
    mics_virtual = Property(desc='microphone geometry')

    _mics_virtual = Instance(MicGeom, desc='internal microphone geometry;internal usage, read only')

    def _get_mics_virtual(self):
        if not self._mics_virtual and self.mics:
            self._mics_virtual = self.mics
        return self._mics_virtual

    def _set_mics_virtual(self, mics_virtual):
        self._mics_virtual = mics_virtual

    # Interpolation method in spatial domain, defaults to linear
    # linear uses numpy linear interpolation
    # spline uses scipy CloughTocher algorithm
    # rbf is scipy radial basis function with multiquadric, cubic and sinc functions
    # idw refers to the inverse distance weighting algorithm
    method = Enum(
        'linear',
        'spline',
        'rbf-multiquadric',
        'rbf-cubic',
        'IDW',
        'custom',
        'sinc',
        desc='method for interpolation used',
    )

    # spatial dimensionality of the array geometry
    array_dimension = Enum('1D', '2D', 'ring', '3D', 'custom', desc='spatial dimensionality of the array geometry')

    # Sampling frequency of output signal, as given by :attr:`source`.
    sample_freq = Delegate('source', 'sample_freq')

    # Number of channels in output.
    num_channels = Property()

    # Number of samples in output, as given by :attr:`source`.
    num_samples = Delegate('source', 'num_samples')

    # Interpolate a point at the origin of the Array geometry
    interp_at_zero = Bool(False)

    # The rotation must be around the z-axis, which means from x to y axis.
    # If the coordinates are not build like that, than this 3x3 orthogonal
    # transformation matrix Q can be used to modify the coordinates.
    # It is assumed that with the modified coordinates the rotation is around the z-axis.
    # The transformation is done via [x,y,z]_mod = Q * [x,y,z]. (default is Identity).
    Q = CArray(dtype=float64, shape=(3, 3), value=identity(3))

    num_IDW = Int(3, desc='number of neighboring microphones, DEFAULT=3')  # noqa: N815

    p_weight = Float(
        2,
        desc='used in interpolation for virtual microphone, weighting power exponent for IDW',
    )

    # Stores the output of :meth:`_virtNewCoord_func`; Read-Only
    _virtNewCoord_func = Property(  # noqa: N815
        depends_on=['mics.digest', 'mics_virtual.digest', 'method', 'array_dimension', 'interp_at_zero'],
    )

    # internal identifier
    digest = Property(
        depends_on=[
            'mics.digest',
            'mics_virtual.digest',
            'source.digest',
            'method',
            'array_dimension',
            'Q',
            'interp_at_zero',
        ],
    )

    def _get_num_channels(self):
        return self.mics_virtual.num_mics

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_virtNewCoord(self):  # noqa N802
        return self._virtNewCoord_func(self.mics.mpos, self.mics_virtual.mpos, self.method, self.array_dimension)

    def sinc_mic(self, r):
        """Modified Sinc function for Radial Basis function approximation."""
        return sinc((r * self.mics_virtual.mpos.shape[1]) / (pi))

    def _virtNewCoord_func(self, mpos, mpos_virt, method, array_dimension):  # noqa N802
        """Core functionality for getting the interpolation.

        Parameters
        ----------
        mpos : float[3, nPhysicalMics]
            The mic positions of the physical (really existing) mics
        mpos_virt : float[3, nVirtualMics]
            The mic positions of the virtual mics
        method : string
            The Interpolation method to use
        array_dimension : string
            The Array Dimensions in cylinder coordinates

        Returns
        -------
        mesh : List[]
            The items of these lists depend on the reduced interpolation dimension of each subarray.
            If the Array is 1D the list items are:
                1. item : float64[nMicsInSpecificSubarray]
                    Ordered positions of the real mics on the new 1d axis,
                    to be used as inputs for numpys interp.
                2. item : int64[nMicsInArray]
                    Indices identifying how the measured pressures must be evaluated, s.t. the
                    entries of the previous item (see last line) correspond to their initial
                    pressure values.
            If the Array is 2D or 3d the list items are:
                1. item : Delaunay mesh object
                    Delaunay mesh (see scipy.spatial.Delaunay) for the specific Array
                2. item : int64[nMicsInArray]
                    same as 1d case, BUT with the difference, that here the rotational periodicity
                    is handled, when constructing the mesh. Therefore, the mesh could have more
                    vertices than the actual Array mics.

        virtNewCoord : float64[3, nVirtualMics]
            Projection of each virtual mic onto its new coordinates. The columns of virtNewCoord
            correspond to [phi, rho, z].

        newCoord : float64[3, nMics]
            Projection of each mic onto its new coordinates. The columns of newCoordinates
            correspond to [phi, rho, z].
        """
        # init positions of virtual mics in cyl coordinates
        nVirtMics = mpos_virt.shape[1]
        virtNewCoord = zeros((3, nVirtMics))
        virtNewCoord.fill(nan)
        # init real positions in cyl coordinates
        nMics = mpos.shape[1]
        newCoord = zeros((3, nMics))
        newCoord.fill(nan)
        # empty mesh object
        mesh = []

        if self.array_dimension == '1D' or self.array_dimension == 'ring':
            # get projections onto new coordinate, for real mics
            projectionOnNewAxis = cartToCyl(mpos, self.Q)[0]
            indReorderHelp = argsort(projectionOnNewAxis)
            mesh.append([projectionOnNewAxis[indReorderHelp], indReorderHelp])

            # new coordinates of real mics
            indReorderHelp = argsort(cartToCyl(mpos, self.Q)[0])
            newCoord = (cartToCyl(mpos, self.Q).T)[indReorderHelp].T

            # and for virtual mics
            virtNewCoord = cartToCyl(mpos_virt)

        elif self.array_dimension == '2D':  # 2d case0
            # get virtual mic projections on new coord system
            virtNewCoord = cartToCyl(mpos_virt, self.Q)

            # new coordinates of real mics
            indReorderHelp = argsort(cartToCyl(mpos, self.Q)[0])
            newCoord = cartToCyl(mpos, self.Q)

            # scipy delauney triangulation
            # Delaunay
            tri = Delaunay(newCoord.T[:, :2], incremental=True)  #

            if self.interp_at_zero:
                # add a point at zero
                tri.add_points(array([[0], [0]]).T)

            # extend mesh with closest boundary points of repeating mesh
            pointsOriginal = arange(tri.points.shape[0])
            hull = tri.convex_hull
            hullPoints = unique(hull)

            addRight = tri.points[hullPoints]
            addRight[:, 0] += 2 * pi
            addLeft = tri.points[hullPoints]
            addLeft[:, 0] -= 2 * pi

            indOrigPoints = concatenate((pointsOriginal, pointsOriginal[hullPoints], pointsOriginal[hullPoints]))
            # add all hull vertices to original mesh and check which of those
            # are actual neighbors of the original array. Cancel out all others.
            tri.add_points(concatenate([addLeft, addRight]))
            indices, indptr = tri.vertex_neighbor_vertices
            hullNeighbor = empty((0), dtype='int32')
            for currHull in hullPoints:
                neighborOfHull = indptr[indices[currHull] : indices[currHull + 1]]
                hullNeighbor = append(hullNeighbor, neighborOfHull)
            hullNeighborUnique = unique(hullNeighbor)
            pointsNew = unique(append(pointsOriginal, hullNeighborUnique))
            tri = Delaunay(tri.points[pointsNew])  # re-meshing
            mesh.append([tri, indOrigPoints[pointsNew]])

        elif self.array_dimension == '3D':  # 3d case
            # get virtual mic projections on new coord system
            virtNewCoord = cartToCyl(mpos_virt, self.Q)
            # get real mic projections on new coord system
            indReorderHelp = argsort(cartToCyl(mpos, self.Q)[0])
            newCoord = cartToCyl(mpos, self.Q)
            # Delaunay
            tri = Delaunay(newCoord.T, incremental=True)  # , incremental=True,qhull_options =  "Qc QJ Q12"

            if self.interp_at_zero:
                # add a point at zero
                tri.add_points(array([[0], [0], [0]]).T)

            # extend mesh with closest boundary points of repeating mesh
            pointsOriginal = arange(tri.points.shape[0])
            hull = tri.convex_hull
            hullPoints = unique(hull)

            addRight = tri.points[hullPoints]
            addRight[:, 0] += 2 * pi
            addLeft = tri.points[hullPoints]
            addLeft[:, 0] -= 2 * pi

            indOrigPoints = concatenate((pointsOriginal, pointsOriginal[hullPoints], pointsOriginal[hullPoints]))
            # add all hull vertices to original mesh and check which of those
            # are actual neighbors of the original array. Cancel out all others.
            tri.add_points(concatenate([addLeft, addRight]))
            indices, indptr = tri.vertex_neighbor_vertices
            hullNeighbor = empty((0), dtype='int32')
            for currHull in hullPoints:
                neighborOfHull = indptr[indices[currHull] : indices[currHull + 1]]
                hullNeighbor = append(hullNeighbor, neighborOfHull)
            hullNeighborUnique = unique(hullNeighbor)
            pointsNew = unique(append(pointsOriginal, hullNeighborUnique))
            tri = Delaunay(tri.points[pointsNew])  # re-meshing
            mesh.append([tri, indOrigPoints[pointsNew]])

        return mesh, virtNewCoord, newCoord

    def _result_core_func(self, p, phi_delay=None, period=None, Q=Q, interp_at_zero=False):  # noqa: N803, ARG002 (see #226)
        """Performs the actual Interpolation.

        Parameters
        ----------
        p : float[num, nMicsReal]
            The pressure field of the yielded sample at real mics.
        phi_delay : empty list (default) or float[num]
            If passed (rotational case), this list contains the angular delay
            of each sample in rad.
        period : None (default) or float
            If periodicity can be assumed (rotational case)
            this parameter contains the periodicity length

        Returns
        -------
        pInterp : float[num, nMicsVirtual]
            The interpolated time data at the virtual mics

        """
        if phi_delay is None:
            phi_delay = []
        # number of time samples
        nTime = p.shape[0]
        # number of virtual mixcs
        nVirtMics = self.mics_virtual.mpos.shape[1]
        # mesh and projection onto polar Coordinates
        meshList, virtNewCoord, newCoord = self._get_virtNewCoord()
        # pressure interpolation init
        pInterp = zeros((nTime, nVirtMics))
        # Coordinates in cartesian CO - for IDW interpolation
        newCoordCart = cylToCart(newCoord)

        if self.interp_at_zero:
            # interpolate point at 0 in Kartesian CO
            interpolater = LinearNDInterpolator(
                cylToCart(newCoord[:, argsort(newCoord[0])])[:2, :].T,
                p[:, (argsort(newCoord[0]))].T,
                fill_value=0,
            )
            pZero = interpolater((0, 0))
            # add the interpolated pressure at origin to pressure channels
            p = concatenate((p, pZero[:, newaxis]), axis=1)

        # helpfunction reordered for reordered pressure values
        pHelp = p[:, meshList[0][1]]

        # Interpolation for 1D Arrays
        if self.array_dimension == '1D' or self.array_dimension == 'ring':
            # for rotation add phi_delay
            if not array_equal(phi_delay, []):
                xInterpHelp = tile(virtNewCoord[0, :], (nTime, 1)) + tile(phi_delay, (virtNewCoord.shape[1], 1)).T
                xInterp = ((xInterpHelp + pi) % (2 * pi)) - pi  #  shifting phi cootrdinate into feasible area [-pi, pi]
            # if no rotation given
            else:
                xInterp = tile(virtNewCoord[0, :], (nTime, 1))
            # get ordered microphone positions in radiant
            x = newCoord[0]
            for cntTime in range(nTime):
                if self.method == 'linear':
                    # numpy 1-d interpolation
                    pInterp[cntTime] = interp(
                        xInterp[cntTime, :],
                        x,
                        pHelp[cntTime, :],
                        period=period,
                        left=nan,
                        right=nan,
                    )

                elif self.method == 'spline':
                    # scipy cubic spline interpolation
                    SplineInterp = CubicSpline(
                        append(x, (2 * pi) + x[0]),
                        append(pHelp[cntTime, :], pHelp[cntTime, :][0]),
                        axis=0,
                        bc_type='periodic',
                        extrapolate=None,
                    )
                    pInterp[cntTime] = SplineInterp(xInterp[cntTime, :])

                elif self.method == 'sinc':
                    # compute using 3-D Rbfs for sinc
                    rbfi = Rbf(
                        x,
                        newCoord[1],
                        newCoord[2],
                        pHelp[cntTime, :],
                        function=self.sinc_mic,
                    )  # radial basis function interpolator instance

                    pInterp[cntTime] = rbfi(xInterp[cntTime, :], virtNewCoord[1], virtNewCoord[2])

                elif self.method == 'rbf-cubic':
                    # compute using 3-D Rbfs with multiquadratics
                    rbfi = Rbf(
                        x,
                        newCoord[1],
                        newCoord[2],
                        pHelp[cntTime, :],
                        function='cubic',
                    )  # radial basis function interpolator instance

                    pInterp[cntTime] = rbfi(xInterp[cntTime, :], virtNewCoord[1], virtNewCoord[2])

        # Interpolation for arbitrary 2D Arrays
        elif self.array_dimension == '2D':
            # check rotation
            if not array_equal(phi_delay, []):
                xInterpHelp = tile(virtNewCoord[0, :], (nTime, 1)) + tile(phi_delay, (virtNewCoord.shape[1], 1)).T
                xInterp = ((xInterpHelp + pi) % (2 * pi)) - pi  # shifting phi cootrdinate into feasible area [-pi, pi]
            else:
                xInterp = tile(virtNewCoord[0, :], (nTime, 1))

            mesh = meshList[0][0]
            for cntTime in range(nTime):
                # points for interpolation
                newPoint = concatenate((xInterp[cntTime, :][:, newaxis], virtNewCoord[1, :][:, newaxis]), axis=1)
                # scipy 1D interpolation
                if self.method == 'linear':
                    interpolater = LinearNDInterpolator(mesh, pHelp[cntTime, :], fill_value=0)
                    pInterp[cntTime] = interpolater(newPoint)

                elif self.method == 'spline':
                    # scipy CloughTocher interpolation
                    f = CloughTocher2DInterpolator(mesh, pHelp[cntTime, :], fill_value=0)
                    pInterp[cntTime] = f(newPoint)

                elif self.method == 'sinc':
                    # compute using 3-D Rbfs for sinc
                    rbfi = Rbf(
                        newCoord[0],
                        newCoord[1],
                        newCoord[2],
                        pHelp[cntTime, : len(newCoord[0])],
                        function=self.sinc_mic,
                    )  # radial basis function interpolator instance

                    pInterp[cntTime] = rbfi(xInterp[cntTime, :], virtNewCoord[1], virtNewCoord[2])

                elif self.method == 'rbf-cubic':
                    # compute using 3-D Rbfs
                    rbfi = Rbf(
                        newCoord[0],
                        newCoord[1],
                        newCoord[2],
                        pHelp[cntTime, : len(newCoord[0])],
                        function='cubic',
                    )  # radial basis function interpolator instance

                    virtshiftcoord = array([xInterp[cntTime, :], virtNewCoord[1], virtNewCoord[2]])
                    pInterp[cntTime] = rbfi(virtshiftcoord[0], virtshiftcoord[1], virtshiftcoord[2])

                elif self.method == 'rbf-multiquadric':
                    # compute using 3-D Rbfs
                    rbfi = Rbf(
                        newCoord[0],
                        newCoord[1],
                        newCoord[2],
                        pHelp[cntTime, : len(newCoord[0])],
                        function='multiquadric',
                    )  # radial basis function interpolator instance

                    virtshiftcoord = array([xInterp[cntTime, :], virtNewCoord[1], virtNewCoord[2]])
                    pInterp[cntTime] = rbfi(virtshiftcoord[0], virtshiftcoord[1], virtshiftcoord[2])
                # using inverse distance weighting
                elif self.method == 'IDW':
                    newPoint2_M = newPoint.T
                    newPoint3_M = append(newPoint2_M, zeros([1, self.num_channels]), axis=0)
                    newPointCart = cylToCart(newPoint3_M)
                    for ind in arange(len(newPoint[:, 0])):
                        newPoint_Rep = tile(newPointCart[:, ind], (len(newPoint[:, 0]), 1)).T
                        subtract = newPoint_Rep - newCoordCart
                        normDistance = norm(subtract, axis=0)
                        index_norm = argsort(normDistance)[: self.num_IDW]
                        pHelpNew = pHelp[cntTime, index_norm]
                        normNew = normDistance[index_norm]
                        if normNew[0] < 1e-3:
                            pInterp[cntTime, ind] = pHelpNew[0]
                        else:
                            wholeD = sum(1 / normNew**self.p_weight)
                            weight = (1 / normNew**self.p_weight) / wholeD
                            pInterp[cntTime, ind] = sum(pHelpNew * weight)

        # Interpolation for arbitrary 3D Arrays
        elif self.array_dimension == '3D':
            # check rotation
            if not array_equal(phi_delay, []):
                xInterpHelp = tile(virtNewCoord[0, :], (nTime, 1)) + tile(phi_delay, (virtNewCoord.shape[1], 1)).T
                xInterp = ((xInterpHelp + pi) % (2 * pi)) - pi  # shifting phi cootrdinate into feasible area [-pi, pi]
            else:
                xInterp = tile(virtNewCoord[0, :], (nTime, 1))

            mesh = meshList[0][0]
            for cntTime in range(nTime):
                # points for interpolation
                newPoint = concatenate((xInterp[cntTime, :][:, newaxis], virtNewCoord[1:, :].T), axis=1)

                if self.method == 'linear':
                    interpolater = LinearNDInterpolator(mesh, pHelp[cntTime, :], fill_value=0)
                    pInterp[cntTime] = interpolater(newPoint)

                elif self.method == 'sinc':
                    # compute using 3-D Rbfs for sinc
                    rbfi = Rbf(
                        newCoord[0],
                        newCoord[1],
                        newCoord[2],
                        pHelp[cntTime, : len(newCoord[0])],
                        function=self.sinc_mic,
                    )  # radial basis function interpolator instance

                    pInterp[cntTime] = rbfi(xInterp[cntTime, :], virtNewCoord[1], virtNewCoord[2])

                elif self.method == 'rbf-cubic':
                    # compute using 3-D Rbfs
                    rbfi = Rbf(
                        newCoord[0],
                        newCoord[1],
                        newCoord[2],
                        pHelp[cntTime, : len(newCoord[0])],
                        function='cubic',
                    )  # radial basis function interpolator instance

                    pInterp[cntTime] = rbfi(xInterp[cntTime, :], virtNewCoord[1], virtNewCoord[2])

                elif self.method == 'rbf-multiquadric':
                    # compute using 3-D Rbfs
                    rbfi = Rbf(
                        newCoord[0],
                        newCoord[1],
                        newCoord[2],
                        pHelp[cntTime, : len(newCoord[0])],
                        function='multiquadric',
                    )  # radial basis function interpolator instance

                    pInterp[cntTime] = rbfi(xInterp[cntTime, :], virtNewCoord[1], virtNewCoord[2])

        # return interpolated pressure values
        return pInterp

    def result(self, num):
        msg = 'result method not implemented yet! Data from source will be passed without transformation.'
        warn(msg, Warning, stacklevel=2)
        yield from self.source.result(num)


class SpatialInterpolatorRotation(SpatialInterpolator):  # pragma: no cover
    """Spatial  Interpolation for rotating sources. Gets samples from :attr:`source`
    and angles from  :attr:`AngleTracker`.Generates output via the generator :meth:`result`.

    """

    # Angle data from AngleTracker class
    angle_source = Instance(AngleTracker)

    # Internal identifier
    digest = Property(
        depends_on=[
            'source.digest',
            'angle_source.digest',
            'mics.digest',
            'mics_virtual.digest',
            'method',
            'array_dimension',
            'Q',
            'interp_at_zero',
        ],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=128):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, :attr:`num_channels`).
            The last block may be shorter than num.

        """
        # period for rotation
        period = 2 * pi
        # get angle
        angle = self.angle_source.angle()
        # counter to track angle position in time for each block
        count = 0
        for timeData in self.source.result(num):
            phi_delay = angle[count : count + num]
            interpVal = self._result_core_func(timeData, phi_delay, period, self.Q, interp_at_zero=False)
            yield interpVal
            count += num


class SpatialInterpolatorConstantRotation(SpatialInterpolator):  # pragma: no cover
    """Spatial linear Interpolation for constantly rotating sources.
    Gets samples from :attr:`source` and generates output via the
    generator :meth:`result`.
    """

    # Rotational speed in rps. Positive, if rotation is around positive z-axis sense,
    # which means from x to y axis.
    rotational_speed = Float(0.0)

    # internal identifier
    digest = Property(
        depends_on=[
            'source.digest',
            'mics.digest',
            'mics_virtual.digest',
            'method',
            'array_dimension',
            'Q',
            'interp_at_zero',
            'rotational_speed',
        ],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=1):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, :attr:`num_channels`).
            The last block may be shorter than num.

        """
        omega = 2 * pi * self.rotational_speed
        period = 2 * pi
        phiOffset = 0.0
        for timeData in self.source.result(num):
            nTime = timeData.shape[0]
            phi_delay = phiOffset + linspace(0, nTime / self.sample_freq * omega, nTime, endpoint=False)
            interpVal = self._result_core_func(timeData, phi_delay, period, self.Q, interp_at_zero=False)
            phiOffset = phi_delay[-1] + omega / self.sample_freq
            yield interpVal


class Mixer(TimeOut):
    """Mixes the signals from several sources."""

    # Data source; :class:`~acoular.base.SamplesGenerator` object.
    source = Instance(SamplesGenerator)

    # List of additional :class:`~acoular.base.SamplesGenerator` objects
    # to be mixed.
    sources = List(Instance(SamplesGenerator, ()))

    # Sampling frequency of the signal as given by :attr:`source`.
    sample_freq = Delegate('source')

    # Number of channels in output as given by :attr:`source`.
    num_channels = Delegate('source')

    # Number of samples in output as given by :attr:`source`.
    num_samples = Delegate('source')

    # internal identifier
    sdigest = Str()

    @observe('sources.items.digest')
    def _set_sourcesdigest(self, event):  # noqa ARG002
        self.sdigest = ldigest(self.sources)

    # internal identifier
    digest = Property(depends_on=['source.digest', 'sdigest'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    def validate_sources(self):
        """Validates if sources fit together."""
        if self.source:
            for s in self.sources:
                if self.sample_freq != s.sample_freq:
                    msg = f'Sample frequency of {s} does not fit'
                    raise ValueError(msg)
                if self.num_channels != s.num_channels:
                    msg = f'Channel count of {s} does not fit'
                    raise ValueError(msg)

    def result(self, num):
        """Python generator that yields the output block-wise.
        The output from the source and those in the list
        sources are being added.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, num_channels).
            The last block may be shorter than num.

        """
        # check whether all sources fit together
        self.validate_sources()

        gens = [i.result(num) for i in self.sources]
        for temp in self.source.result(num):
            sh = temp.shape[0]
            for g in gens:
                try:
                    temp1 = next(g)
                except StopIteration:
                    return
                if temp.shape[0] > temp1.shape[0]:
                    temp = temp[: temp1.shape[0]]
                temp += temp1[: temp.shape[0]]
            yield temp
            if sh > temp.shape[0]:
                break


class TimePower(TimeOut):
    """Calculates time-depended power of the signal."""

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Squared output of source.
            Yields samples in blocks of shape (num, num_channels).
            The last block may be shorter than num.

        """
        for temp in self.source.result(num):
            yield temp * temp


class TimeCumAverage(TimeOut):
    """Calculates cumulative average of the signal, useful for Leq."""

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Cumulative average of the output of source.
            Yields samples in blocks of shape (num, num_channels).
            The last block may be shorter than num.

        """
        count = (arange(num) + 1)[:, newaxis]
        for i, temp in enumerate(self.source.result(num)):
            ns, nc = temp.shape
            if not i:
                accu = zeros((1, nc))
            temp = (accu * (count[0] - 1) + cumsum(temp, axis=0)) / count[:ns]
            accu = temp[-1]
            count += ns
            yield temp


class TimeReverse(TimeOut):
    """Calculates the time-reversed signal of a source."""

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Yields samples in blocks of shape (num, num_channels).
            Time-reversed output of source.
            The last block may be shorter than num.

        """
        result_list = []
        result_list.extend(self.source.result(num))
        temp = empty_like(result_list[0])
        h = result_list.pop()
        nsh = h.shape[0]
        temp[:nsh] = h[::-1]
        for h in result_list[::-1]:
            temp[nsh:] = h[: nsh - 1 : -1]
            yield temp
            temp[:nsh] = h[nsh - 1 :: -1]
        yield temp[:nsh]


class Filter(TimeOut):
    """Abstract base class for IIR filters based on scipy lfilter
    implements a filter with coefficients that may be changed
    during processing.

    Should not be instantiated by itself.
    """

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Filter coefficients
    sos = Property()

    def _get_sos(self):
        return tf2sos([1], [1])

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, num_channels).
            Delivers the bandpass filtered output of source.
            The last block may be shorter than num.

        """
        sos = self.sos
        zi = zeros((sos.shape[0], 2, self.source.num_channels))
        for block in self.source.result(num):
            sos = self.sos  # this line is useful in case of changes
            # to self.sos during generator lifetime
            block, zi = sosfilt(sos, block, axis=0, zi=zi)
            yield block


class FiltOctave(Filter):
    """Octave or third-octave filter (causal, non-zero phase delay)."""

    # Band center frequency; defaults to 1000.
    band = Float(1000.0, desc='band center frequency')

    # Octave fraction: 'Octave' or 'Third octave'; defaults to 'Octave'.
    fraction = Map({'Octave': 1, 'Third octave': 3}, default_value='Octave', desc='fraction of octave')

    # Filter order
    order = Int(3, desc='IIR filter order')

    sos = Property(depends_on=['band', 'fraction', 'source.digest', 'order'])

    # internal identifier
    digest = Property(depends_on=['source.digest', 'band', 'fraction', 'order'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_sos(self):
        # filter design
        fs = self.sample_freq
        # adjust filter edge frequencies for correct power bandwidth (see ANSI 1.11 1987
        # and Kalb,J.T.: "A thirty channel real time audio analyzer and its applications",
        # PhD Thesis: Georgia Inst. of Techn., 1975
        beta = pi / (2 * self.order)
        alpha = pow(2.0, 1.0 / (2.0 * self.fraction_))
        beta = 2 * beta / sin(beta) / (alpha - 1 / alpha)
        alpha = (1 + sqrt(1 + beta * beta)) / beta
        fr = 2 * self.band / fs
        if fr > 1 / sqrt(2):
            msg = f'band frequency too high:{self.band:f},{fs:f}'
            raise ValueError(msg)
        om1 = fr / alpha
        om2 = fr * alpha
        return butter(self.order, [om1, om2], 'bandpass', output='sos')


class FiltFiltOctave(FiltOctave):
    """Octave or third-octave filter with zero phase delay.

    This filter can be applied on time signals.
    It requires large amounts of memory!
    """

    # Filter order (applied for forward filter and backward filter)
    order = Int(2, desc='IIR filter half order')

    # internal identifier
    digest = Property(depends_on=['source.digest', 'band', 'fraction', 'order'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_sos(self):
        # filter design
        fs = self.sample_freq
        # adjust filter edge frequencies for correct power bandwidth (see FiltOctave)
        beta = pi / (2 * self.order)
        alpha = pow(2.0, 1.0 / (2.0 * self.fraction_))
        beta = 2 * beta / sin(beta) / (alpha - 1 / alpha)
        alpha = (1 + sqrt(1 + beta * beta)) / beta
        # additional bandwidth correction for double-pass
        alpha = alpha * {6: 1.01, 5: 1.012, 4: 1.016, 3: 1.022, 2: 1.036, 1: 1.083}.get(self.order, 1.0) ** (
            3 / self.fraction_
        )
        fr = 2 * self.band / fs
        if fr > 1 / sqrt(2):
            msg = f'band frequency too high:{self.band:f},{fs:f}'
            raise ValueError(msg)
        om1 = fr / alpha
        om2 = fr * alpha
        return butter(self.order, [om1, om2], 'bandpass', output='sos')

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, num_channels).
            Delivers the zero-phase bandpass filtered output of source.
            The last block may be shorter than num.

        """
        sos = self.sos
        data = empty((self.source.num_samples, self.source.num_channels))
        j = 0
        for block in self.source.result(num):
            ns, nc = block.shape
            data[j : j + ns] = block
            j += ns
        # filter one channel at a time to save memory
        for j in range(self.source.num_channels):
            data[:, j] = sosfiltfilt(sos, data[:, j])
        j = 0
        ns = data.shape[0]
        while j < ns:
            yield data[j : j + num]
            j += num


class TimeExpAverage(Filter):
    """Computes exponential averaging according to IEC 61672-1
    time constant: F -> 125 ms, S -> 1 s
    I (non-standard) -> 35 ms.
    """

    # time weighting
    weight = Map({'F': 0.125, 'S': 1.0, 'I': 0.035}, default_value='F', desc='time weighting')

    sos = Property(depends_on=['weight', 'source.digest'])

    # internal identifier
    digest = Property(depends_on=['source.digest', 'weight'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_sos(self):
        alpha = 1 - exp(-1 / self.weight_ / self.sample_freq)
        a = [1, alpha - 1]
        b = [alpha]
        return tf2sos(b, a)


class FiltFreqWeight(Filter):
    """Frequency weighting filter according to IEC 61672."""

    # weighting characteristics
    weight = Enum('A', 'C', 'Z', desc='frequency weighting')

    sos = Property(depends_on=['weight', 'source.digest'])

    # internal identifier
    digest = Property(depends_on=['source.digest', 'weight'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_sos(self):
        # s domain coefficients
        f1 = 20.598997
        f2 = 107.65265
        f3 = 737.86223
        f4 = 12194.217
        a = polymul([1, 4 * pi * f4, (2 * pi * f4) ** 2], [1, 4 * pi * f1, (2 * pi * f1) ** 2])
        if self.weight == 'A':
            a = polymul(polymul(a, [1, 2 * pi * f3]), [1, 2 * pi * f2])
            b = [(2 * pi * f4) ** 2 * 10 ** (1.9997 / 20), 0, 0, 0, 0]
            b, a = bilinear(b, a, self.sample_freq)
        elif self.weight == 'C':
            b = [(2 * pi * f4) ** 2 * 10 ** (0.0619 / 20), 0, 0]
            b, a = bilinear(b, a, self.sample_freq)
            b = append(b, zeros(2))  # make 6th order
            a = append(a, zeros(2))
        else:
            b = zeros(7)
            b[0] = 1.0
            a = b  # 6th order flat response
        return tf2sos(b, a)


@deprecated_alias({'numbands': 'num_bands'}, read_only=True)
class FilterBank(TimeOut):
    """Abstract base class for IIR filter banks based on scipy lfilter
    implements a bank of parallel filters.

    Should not be instantiated by itself.
    """

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # List of filter coefficients for all filters
    sos = Property()

    # List of labels for bands
    bands = Property()

    # Number of bands
    num_bands = Property()

    # Number of bands
    num_channels = Property()

    @abstractmethod
    def _get_sos(self):
        """Returns a list of second order section coefficients."""

    @abstractmethod
    def _get_bands(self):
        """Returns a list of labels for the bands."""

    @abstractmethod
    def _get_num_bands(self):
        """Returns the number of bands."""

    def _get_num_channels(self):
        return self.num_bands * self.source.num_channels

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, num_channels).
            Delivers the bandpass filtered output of source.
            The last block may be shorter than num.

        """
        numbands = self.num_bands
        snumch = self.source.num_channels
        sos = self.sos
        zi = [zeros((sos[0].shape[0], 2, snumch)) for _ in range(numbands)]
        res = zeros((num, self.num_channels), dtype='float')
        for block in self.source.result(num):
            for i in range(numbands):
                res[:, i * snumch : (i + 1) * snumch], zi[i] = sosfilt(sos[i], block, axis=0, zi=zi[i])
            yield res


class OctaveFilterBank(FilterBank):
    """Octave or third-octave filter bank."""

    # Lowest band center frequency index; defaults to 21 (=125 Hz).
    lband = Int(21, desc='lowest band center frequency index')

    # Lowest band center frequency index + 1; defaults to 40 (=8000 Hz).
    hband = Int(40, desc='lowest band center frequency index')

    # Octave fraction: 'Octave' or 'Third octave'; defaults to 'Octave'.
    fraction = Map({'Octave': 1, 'Third octave': 3}, default_value='Octave', desc='fraction of octave')

    # List of filter coefficients for all filters
    ba = Property(depends_on=['lband', 'hband', 'fraction', 'source.digest'])

    # List of labels for bands
    bands = Property(depends_on=['lband', 'hband', 'fraction'])

    # Number of bands
    num_bands = Property(depends_on=['lband', 'hband', 'fraction'])

    # internal identifier
    digest = Property(depends_on=['source.digest', 'lband', 'hband', 'fraction', 'order'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_bands(self):
        return [10 ** (i / 10) for i in range(self.lband, self.hband, 4 - self.fraction_)]

    @cached_property
    def _get_num_bands(self):
        return len(self.bands)

    @cached_property
    def _get_sos(self):
        of = FiltOctave(source=self.source, fraction=self.fraction)
        sos = []
        for i in range(self.lband, self.hband, 4 - self.fraction_):
            of.band = 10 ** (i / 10)
            sos_ = of.sos
            sos.append(sos_)
        return sos


@deprecated_alias({'name': 'file'})
class WriteWAV(TimeOut):
    """Saves time signal from one or more channels as mono/stereo/multi-channel
    `*.wav` file.
    """

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Name of the file to be saved. If none is given, the name will be
    # automatically generated from the sources.
    file = File(filter=['*.wav'], desc='name of wave file')

    # Basename for cache, readonly.
    basename = Property(depends_on=['digest'])

    # Channel(s) to save. List can only contain one or two channels.
    channels = List(int, desc='channel to save')

    # internal identifier
    digest = Property(depends_on=['source.digest', 'channels'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_basename(self):
        warn(
            (
                f'The basename attribute of a {self.__class__.__name__} object is deprecated'
                ' and will be removed in a future release!'
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return find_basename(self.source)

    def save(self):
        """Saves source output to one- or multiple-channel `*.wav` file."""
        nc = len(self.channels)
        if nc == 0:
            msg = 'No channels given for output.'
            raise ValueError(msg)
        if nc > 2:
            warn(f'More than two channels given for output, exported file will have {nc:d} channels', stacklevel=1)
        if self.file == '':
            name = self.basename
            for nr in self.channels:
                name += f'{nr:d}'
            name += '.wav'
        else:
            name = self.file
        with wave.open(name, 'w') as wf:
            wf.setnchannels(nc)
            wf.setsampwidth(2)
            wf.setframerate(self.source.sample_freq)
            wf.setnframes(self.source.num_samples)
            mx = 0.0
            ind = array(self.channels)
            for data in self.source.result(1024):
                mx = max(abs(data[:, ind]).max(), mx)
            scale = 0.9 * 2**15 / mx
            for data in self.source.result(1024):
                wf.writeframesraw(array(data[:, ind] * scale, dtype=int16).tostring())

    def result(self, num):
        msg = 'result method not implemented yet! Data from source will be passed without transformation.'
        warn(msg, Warning, stacklevel=2)
        yield from self.source.result(num)


@deprecated_alias({'name': 'file', 'numsamples_write': 'num_samples_write', 'writeflag': 'write_flag'})
class WriteH5(TimeOut):
    """Saves time signal as `*.h5` file."""

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Name of the file to be saved. If none is given, the name will be
    # automatically generated from a time stamp.
    file = File(filter=['*.h5'], desc='name of data file')

    # Number of samples to write to file by `result` method.
    # defaults to -1 (write as long as source yields data).
    num_samples_write = Int(-1)

    # flag that can be raised to stop file writing
    write_flag = Bool(True)

    # internal identifier
    digest = Property(depends_on=['source.digest'])

    # The floating-number-precision of entries of H5 File corresponding
    # to numpy dtypes. Default is 32 bit.
    precision = Enum('float32', 'float64', desc='precision of H5 File')

    # Metadata to be stored in HDF5 file object
    metadata = Dict(desc='metadata to be stored in .h5 file')

    @cached_property
    def _get_digest(self):
        return digest(self)

    def create_filename(self):
        if self.file == '':
            name = datetime.now(tz=timezone.utc).isoformat('_').replace(':', '-').replace('.', '_')
            self.file = path.join(config.td_dir, name + '.h5')

    def get_initialized_file(self):
        file = _get_h5file_class()
        self.create_filename()
        f5h = file(self.file, mode='w')
        f5h.create_extendable_array('time_data', (0, self.num_channels), self.precision)
        ac = f5h.get_data_by_reference('time_data')
        f5h.set_node_attribute(ac, 'sample_freq', self.sample_freq)
        self.add_metadata(f5h)
        return f5h

    def save(self):
        """Saves source output to `*.h5` file."""
        f5h = self.get_initialized_file()
        ac = f5h.get_data_by_reference('time_data')
        for data in self.source.result(4096):
            f5h.append_data(ac, data)
        f5h.close()

    def add_metadata(self, f5h):
        """Adds metadata to .h5 file."""
        nitems = len(self.metadata.items())
        if nitems > 0:
            f5h.create_new_group('metadata', '/')
            for key, value in self.metadata.items():
                if isinstance(value, str):
                    value = array(value, dtype='S')
                f5h.create_array('/metadata', key, value)

    def result(self, num):
        """Python generator that saves source output to `*.h5` file and
        yields the source output block-wise.


        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, num_channels).
            The last block may be shorter than num.
            Echos the source output, but reads it from cache
            when available and prevents unnecessary recalculation.

        """
        self.write_flag = True
        f5h = self.get_initialized_file()
        ac = f5h.get_data_by_reference('time_data')
        scount = 0
        stotal = self.num_samples_write
        source_gen = self.source.result(num)
        while self.write_flag:
            sleft = stotal - scount
            if stotal != -1 and sleft > 0:
                anz = min(num, sleft)
            elif stotal == -1:
                anz = num
            else:
                break
            try:
                data = next(source_gen)
            except StopIteration:
                break
            f5h.append_data(ac, data[:anz])
            f5h.flush()
            yield data
            scount += anz
        f5h.close()


class TimeConvolve(TimeOut):
    """Fast frequency domain convolution with the Uniformly partitioned overlap-save method (UPOLS).

    See :cite:`Wefers2015` for details.
    """

    # Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Convolution kernel in the time domain. The second dimension of the kernel array has to be
    # either 1 or match :attr:`~SamplesGenerator.num_channels`. If only a single kernel is supplied,
    # it is applied to all channels.
    kernel = CArray(dtype=float, desc='Convolution kernel.')

    _block_size = Int(desc='Block size')

    _kernel_blocks = Property(
        depends_on=['kernel', '_block_size'],
        desc='Frequency domain Kernel blocks',
    )

    # internal identifier
    digest = Property(depends_on=['source.digest', 'kernel'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    def _validate_kernel(self):
        # reshape kernel for broadcasting
        if self.kernel.ndim == 1:
            self.kernel = self.kernel.reshape([-1, 1])
            return
        # check dimensionality
        if self.kernel.ndim > 2:
            msg = 'Only one or two dimensional kernels accepted.'
            raise ValueError(msg)
        # check if number of kernels matches num_channels
        if self.kernel.shape[1] not in (1, self.source.num_channels):
            msg = 'Number of kernels must be either `num_channels` or one.'
            raise ValueError(msg)

    # compute the rfft of the kernel blockwise
    @cached_property
    def _get__kernel_blocks(self):
        [L, N] = self.kernel.shape
        num = self._block_size
        P = int(ceil(L / num))
        trim = num * (P - 1)
        blocks = zeros([P, num + 1, N], dtype='complex128')

        if P > 1:
            for i, block in enumerate(split(self.kernel[:trim], P - 1, axis=0)):
                blocks[i] = rfft(concatenate([block, zeros([num, N])], axis=0), axis=0)

        blocks[-1] = rfft(
            concatenate([self.kernel[trim:], zeros([2 * num - L + trim, N])], axis=0),
            axis=0,
        )
        return blocks

    def result(self, num=128):
        """Python generator that yields the output block-wise.
        The source output is convolved with the kernel.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, num_channels).
            The last block may be shorter than num.

        """
        self._validate_kernel()
        # initialize variables
        self._block_size = num
        L = self.kernel.shape[0]
        N = self.source.num_channels
        M = self.source.num_samples
        numblocks_kernel = int(ceil(L / num))  # number of kernel blocks
        Q = int(ceil(M / num))  # number of signal blocks
        R = int(ceil((L + M - 1) / num))  # number of output blocks
        last_size = (L + M - 1) % num  # size of final block

        idx = 0
        fdl = zeros([numblocks_kernel, num + 1, N], dtype='complex128')
        buff = zeros([2 * num, N])  # time-domain input buffer
        spec_sum = zeros([num + 1, N], dtype='complex128')

        signal_blocks = self.source.result(num)
        temp = next(signal_blocks)
        buff[num : num + temp.shape[0]] = temp  # append new time-data

        # for very short signals, we are already done
        if R == 1:
            _append_to_fdl(fdl, idx, numblocks_kernel, rfft(buff, axis=0))
            spec_sum = _spectral_sum(spec_sum, fdl, self._kernel_blocks)
            # truncate s.t. total length is L+M-1 (like numpy convolve w/ mode="full")
            yield irfft(spec_sum, axis=0)[num : last_size + num]
            return

        # stream processing of source signal
        for temp in signal_blocks:
            _append_to_fdl(fdl, idx, numblocks_kernel, rfft(buff, axis=0))
            spec_sum = _spectral_sum(spec_sum, fdl, self._kernel_blocks)
            yield irfft(spec_sum, axis=0)[num:]
            buff = concatenate(
                [buff[num:], zeros([num, N])],
                axis=0,
            )  # shift input buffer to the left
            buff[num : num + temp.shape[0]] = temp  # append new time-data

        for _ in range(R - Q):
            _append_to_fdl(fdl, idx, numblocks_kernel, rfft(buff, axis=0))
            spec_sum = _spectral_sum(spec_sum, fdl, self._kernel_blocks)
            yield irfft(spec_sum, axis=0)[num:]
            buff = concatenate(
                [buff[num:], zeros([num, N])],
                axis=0,
            )  # shift input buffer to the left

        _append_to_fdl(fdl, idx, numblocks_kernel, rfft(buff, axis=0))
        spec_sum = _spectral_sum(spec_sum, fdl, self._kernel_blocks)
        # truncate s.t. total length is L+M-1 (like numpy convolve w/ mode="full")
        yield irfft(spec_sum, axis=0)[num : last_size + num]


@nb.jit(nopython=True, cache=True)
def _append_to_fdl(fdl, idx, numblocks_kernel, buff):  # pragma: no cover
    fdl[idx] = buff
    idx = int(idx + 1 % numblocks_kernel)


@nb.jit(nopython=True, cache=True)
def _spectral_sum(out, fdl, kb):  # pragma: no cover
    P, B, N = kb.shape
    for n in range(N):
        for b in range(B):
            out[b, n] = 0
            for i in range(P):
                out[b, n] += fdl[i, b, n] * kb[i, b, n]

    return out


class MaskedTimeInOut(MaskedTimeOut):
    """Signal processing block for channel and sample selection.

    .. deprecated:: 24.10
        Using :class:`~acoular.tprocess.MaskedTimeInOut` is deprecated and will be removed in
        Acoular version 25.07. Use :class:`~acoular.tprocess.MaskedTimeOut` instead.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn(
            'Using MaskedTimeInOut is deprecated and will be removed in Acoular version 25.07. \
            Use class MaskedTimeOut instead.',
            DeprecationWarning,
            stacklevel=2,
        )



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements beamformers in the time domain.

.. autosummary::
    :toctree: generated/

    BeamformerTime
    BeamformerTimeTraj
    BeamformerTimeSq
    BeamformerTimeSqTraj
    BeamformerCleant
    BeamformerCleantTraj
    BeamformerCleantSq
    BeamformerCleantSqTraj
    IntegratorSectorTime
"""

# imports from other packages

from numpy import (
    arange,
    argmax,
    array,
    ceil,
    dot,
    empty,
    float32,
    float64,
    histogram,
    int32,
    int64,
    interp,
    isscalar,
    newaxis,
    r_,
    s_,
    sqrt,
    sum,  # noqa: A004
    unique,
    where,
    zeros,
)
from scipy.linalg import norm
from traits.api import Bool, CArray, Enum, Float, Instance, Int, List, Map, Property, Range, cached_property

from .base import SamplesGenerator, TimeOut
from .fbeamform import SteeringVector
from .grids import RectGrid

# acoular imports
from .internal import digest
from .process import SamplesBuffer
from .tfastfuncs import _delayandsum4, _delayandsum5, _delays
from .trajectory import Trajectory


def const_power_weight(bf):
    """Internal helper function for :class:`BeamformerTime`.

    Provides microphone weighting
    to make the power per unit area of the
    microphone array geometry constant.

    Parameters
    ----------
    bf: :class:`BeamformerTime` object


    Returns
    -------
    array of floats
        The weight factors.
    """
    r = bf.steer.env._r(zeros((3, 1)), bf.steer.mics.pos)  # distances to center
    # round the relative distances to one decimal place
    r = (r / r.max()).round(decimals=1)
    ru, ind = unique(r, return_inverse=True)
    ru = (ru[1:] + ru[:-1]) / 2
    count, bins = histogram(r, r_[0, ru, 1.5 * r.max() - 0.5 * ru[-1]])
    bins *= bins
    weights = sqrt((bins[1:] - bins[:-1]) / count)
    weights /= weights.mean()
    return weights[ind]


# possible choices for spatial weights
possible_weights = {'none': None, 'power': const_power_weight}


class BeamformerTime(TimeOut):
    """Provides a basic time domain beamformer with time signal output
    for a spatially fixed grid.
    """

    #: Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Instance of :class:`~acoular.fbeamform.SteeringVector` or its derived classes
    # that contains information about the steering vector. This is a private trait.
    # Do not set this directly, use `steer` trait instead.
    steer = Instance(SteeringVector, args=())

    #: Number of channels in output (=number of grid points).
    num_channels = Property()

    #: Spatial weighting function.
    weights = Map(possible_weights, default_value='none', desc='spatial weighting function')
    # (from timedomain.possible_weights)

    # internal identifier
    digest = Property(
        depends_on=['steer.digest', 'source.digest', 'weights'],
    )

    def _get_num_channels(self):
        return self.steer.grid.size

    def _get_digest(self):
        return digest(self)

    def _get_weights(self):
        return self.weights_(self)[newaxis] if self.weights_ else 1.0

    def result(self, num=2048):
        """Python generator that yields the time-domain beamformer output.

        The output time signal starts for source signals that were emitted from
        the :class:`~acoular.grids.Grid` at `t=0`.

        Parameters
        ----------
        num : int
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). Defaults to 2048.

        Yields
        ------
        numpy.ndarray
            Samples in blocks of shape (num, :attr:`~BeamformerTime.num_channels`).
                :attr:`~BeamformerTime.num_channels` is usually very \
                large (number of grid points).
                The last block returned by the generator may be shorter than num.
        """
        # initialize values
        steer_func = self.steer._steer_funcs_time[self.steer.steer_type]
        fdtype = float64
        idtype = int64
        num_mics = self.steer.mics.num_mics
        n_index = arange(0, num + 1)[:, newaxis]
        c = self.steer.env.c / self.source.sample_freq
        amp = empty((1, self.steer.grid.size, num_mics), dtype=fdtype)
        # delays = empty((1,self.steer.grid.size,num_mics),dtype=fdtype)
        d_index = empty((1, self.steer.grid.size, num_mics), dtype=idtype)
        d_interp2 = empty((1, self.steer.grid.size, num_mics), dtype=fdtype)
        steer_func(self.steer.rm[newaxis, :, :], self.steer.r0[newaxis, :], amp)
        _delays(self.steer.rm[newaxis, :, :], c, d_interp2, d_index)
        amp.shape = amp.shape[1:]
        # delays.shape = delays.shape[1:]
        d_index.shape = d_index.shape[1:]
        d_interp2.shape = d_interp2.shape[1:]
        max_sample_delay = int((self.steer.rm / c).max()) + 2
        weights = self._get_weights()

        buffer = SamplesBuffer(
            source=self.source,
            length=int(ceil((num + max_sample_delay) / num)) * num,
            result_num=num + max_sample_delay,
            shift_index_by='num',
            dtype=fdtype,
        )
        for p_res in buffer.result(num):
            p_res *= weights
            if p_res.shape[0] < buffer.result_num:  # last block shorter
                num = p_res.shape[0] - max_sample_delay
                n_index = arange(0, num + 1)[:, newaxis]
            # init step
            Phi, autopow = self._delay_and_sum(num, p_res, d_interp2, d_index, amp)
            if 'Cleant' not in self.__class__.__name__:
                if 'Sq' not in self.__class__.__name__:
                    yield Phi[:num]
                elif self.r_diag:
                    yield (Phi[:num] ** 2 - autopow[:num]).clip(min=0)
                else:
                    yield Phi[:num] ** 2
            else:
                p_res_copy = p_res.copy()
                Gamma = zeros(Phi.shape)
                Gamma_autopow = zeros(Phi.shape)
                J = 0
                # deconvolution
                while self.n_iter > J:
                    # print(f"start clean iteration {J+1} of max {self.n_iter}")
                    powPhi = (Phi[:num] ** 2 - autopow).sum(0).clip(min=0) if self.r_diag else (Phi[:num] ** 2).sum(0)
                    imax = argmax(powPhi)
                    t_float = d_interp2[imax] + d_index[imax] + n_index
                    t_ind = t_float.astype(int64)
                    for m in range(num_mics):
                        p_res_copy[t_ind[: num + 1, m], m] -= self.damp * interp(
                            t_ind[: num + 1, m],
                            t_float[:num, m],
                            Phi[:num, imax] * self.steer.r0[imax] / self.steer.rm[imax, m],
                        )
                    nextPhi, nextAutopow = self._delay_and_sum(num, p_res_copy, d_interp2, d_index, amp)
                    if self.r_diag:
                        pownextPhi = (nextPhi[:num] ** 2 - nextAutopow).sum(0).clip(min=0)
                    else:
                        pownextPhi = (nextPhi[:num] ** 2).sum(0)
                    # print(f"total signal power: {powPhi.sum()}")
                    if pownextPhi.sum() < powPhi.sum():  # stopping criterion
                        Gamma[:num, imax] += self.damp * Phi[:num, imax]
                        Gamma_autopow[:num, imax] = autopow[:num, imax].copy()
                        Phi = nextPhi
                        autopow = nextAutopow
                        # print(f"clean max: {L_p((Gamma**2).sum(0)/num).max()} dB")
                        J += 1
                    else:
                        break
                if 'Sq' not in self.__class__.__name__:
                    yield Gamma[:num]
                elif self.r_diag:
                    yield Gamma[:num] ** 2 - (self.damp**2) * Gamma_autopow[:num]
                else:
                    yield Gamma[:num] ** 2

    def _delay_and_sum(self, num, p_res, d_interp2, d_index, amp):
        """Standard delay-and-sum method."""
        result = empty((num, self.steer.grid.size), dtype=float)  # output array
        autopow = empty((num, self.steer.grid.size), dtype=float)  # output array
        _delayandsum4(p_res, d_index, d_interp2, amp, result, autopow)
        return result, autopow


class BeamformerTimeSq(BeamformerTime):
    """Provides a time domain beamformer with time-dependend
    power signal output and possible autopower removal
    for a spatially fixed grid.
    """

    #: Boolean flag, if 'True' (default), the main diagonal is removed before beamforming.
    r_diag = Bool(True, desc='removal of diagonal')

    # internal identifier
    digest = Property(
        depends_on=['steer.digest', 'source.digest', 'r_diag', 'weights'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=2048):
        """Python generator that yields the **squared** time-domain beamformer output.

        The squared output time signal starts for source signals that were emitted from
        the :class:`~acoular.grids.Grid` at `t=0`.

        Parameters
        ----------
        num : int
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). Defaults to 2048.

        Yields
        ------
        numpy.ndarray
            Samples in blocks of shape (num, :attr:`~BeamformerTime.num_channels`).
                :attr:`~BeamformerTime.num_channels` is usually very \
                large (number of grid points).
                The last block returned by the generator may be shorter than num.
        """
        return super().result(num)


class BeamformerTimeTraj(BeamformerTime):
    """Provides a basic time domain beamformer with time signal output
    for a grid moving along a trajectory.
    """

    #: :class:`~acoular.trajectory.Trajectory` or derived object.
    #: Start time is assumed to be the same as for the samples.
    trajectory = Instance(Trajectory, desc='trajectory of the grid center')

    #: Reference vector, perpendicular to the y-axis of moving grid.
    rvec = CArray(dtype=float, shape=(3,), value=array((0, 0, 0)), desc='reference vector')

    #: Considering of convective amplification in beamforming formula.
    conv_amp = Bool(False, desc='determines if convective amplification of source is considered')

    #: Floating point and integer precision
    precision = Enum(64, 32, desc='numeric precision')

    # internal identifier
    digest = Property(
        depends_on=[
            'steer.digest',
            'source.digest',
            'weights',
            'precision',
            'rvec',
            'conv_amp',
            'trajectory.digest',
        ],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def _get_moving_gpos(self):
        """Python generator that yields the moving grid coordinates samplewise."""

        def cross(a, b):
            """Cross product for fast computation
            because numpy.cross is ultra slow in this case.
            """
            return array([a[1] * b[2] - a[2] * b[1], a[2] * b[0] - a[0] * b[2], a[0] * b[1] - a[1] * b[0]])

        start_t = 0.0
        gpos = self.steer.grid.pos
        trajg = self.trajectory.traj(start_t, delta_t=1 / self.source.sample_freq)
        trajg1 = self.trajectory.traj(start_t, delta_t=1 / self.source.sample_freq, der=1)
        rflag = (self.rvec == 0).all()  # flag translation vs. rotation
        if rflag:
            for g in trajg:
                # grid is only translated, not rotated
                tpos = gpos + array(g)[:, newaxis]
                yield tpos
        else:
            for g, g1 in zip(trajg, trajg1):
                # grid is both translated and rotated
                loc = array(g)  # translation array([0., 0.4, 1.])
                dx = array(g1)  # direction vector (new x-axis)
                dy = cross(self.rvec, dx)  # new y-axis
                dz = cross(dx, dy)  # new z-axis
                RM = array((dx, dy, dz)).T  # rotation matrix
                RM /= sqrt((RM * RM).sum(0))  # column normalized
                tpos = dot(RM, gpos) + loc[:, newaxis]  # rotation+translation
                #                print(loc[:])
                yield tpos

    def _get_macostheta(self, g1, tpos, rm):
        vvec = array(g1)  # velocity vector
        ma = norm(vvec) / self.steer.env.c  # machnumber
        fdv = (vvec / sqrt((vvec * vvec).sum()))[:, newaxis]  # unit vecor velocity
        mpos = self.steer.mics.pos[:, newaxis, :]
        rmv = tpos[:, :, newaxis] - mpos
        return (ma * sum(rmv.reshape((3, -1)) * fdv, 0) / rm.reshape(-1)).reshape(rm.shape)

    def get_r0(self, tpos):
        if isscalar(self.steer.ref) and self.steer.ref > 0:
            return self.steer.ref  # full((self.steer.grid.size,), self.steer.ref)
        return self.steer.env._r(tpos)

    def result(self, num=2048):
        """Python generator that yields the time-domain beamformer output.

        The output time signal starts for source signals that were emitted from
        the :class:`~acoular.grids.Grid` at `t=0`.

        Parameters
        ----------
        num : int
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). Defaults to 2048.

        Yields
        ------
        numpy.ndarray
            Samples in blocks of shape (num, :attr:`~BeamformerTime.num_channels`).
                :attr:`~BeamformerTime.num_channels` is usually very \
                large (number of grid points).
                The last block returned by the generator may be shorter than num.
        """
        # initialize values
        if self.precision == 64:
            fdtype = float64
            idtype = int64
        else:
            fdtype = float32
            idtype = int32
        w = self._get_weights()
        c = self.steer.env.c / self.source.sample_freq
        num_mics = self.steer.mics.num_mics
        mpos = self.steer.mics.pos.astype(fdtype)
        m_index = arange(num_mics, dtype=idtype)
        n_index = arange(num, dtype=idtype)[:, newaxis]
        blockrm = empty((num, self.steer.grid.size, num_mics), dtype=fdtype)
        blockrmconv = empty((num, self.steer.grid.size, num_mics), dtype=fdtype)
        amp = empty((num, self.steer.grid.size, num_mics), dtype=fdtype)
        # delays = empty((num,self.steer.grid.size,num_mics),dtype=fdtype)
        d_index = empty((num, self.steer.grid.size, num_mics), dtype=idtype)
        d_interp2 = empty((num, self.steer.grid.size, num_mics), dtype=fdtype)
        blockr0 = empty((num, self.steer.grid.size), dtype=fdtype)
        movgpos = self._get_moving_gpos()  # create moving grid pos generator
        movgspeed = self.trajectory.traj(0.0, delta_t=1 / self.source.sample_freq, der=1)
        weights = self._get_weights()

        # preliminary implementation of different steering vectors
        steer_func = self.steer._steer_funcs_time[self.steer.steer_type]

        # start processing
        flag = True
        buffer = SamplesBuffer(source=self.source, length=num * 2, shift_index_by='num', dtype=fdtype)
        buffered_result = buffer.result(num)
        while flag:
            for i in range(num):
                tpos = next(movgpos).astype(fdtype)
                rm = self.steer.env._r(tpos, mpos)  # .astype(fdtype)
                blockr0[i, :] = self.get_r0(tpos)
                blockrm[i, :, :] = rm
                if self.conv_amp:
                    ht = next(movgspeed)
                    blockrmconv[i, :, :] = rm * (1 - self._get_macostheta(ht, tpos, rm)) ** 2
            if self.conv_amp:
                steer_func(blockrmconv, blockr0, amp)
            else:
                steer_func(blockrm, blockr0, amp)
            _delays(blockrm, c, d_interp2, d_index)
            max_sample_delay = (d_index.max((1, 2)) + 2).max()  # + because of interpolation
            buffer.result_num = num + max_sample_delay

            try:
                time_block = next(buffered_result)
                time_block *= weights
            except StopIteration:
                break
            if time_block.shape[0] < buffer.result_num:  # last block shorter
                num = sum((d_index.max((1, 2)) + 1 + arange(0, num)) < time_block.shape[0])
                n_index = arange(num, dtype=idtype)[:, newaxis]
                flag = False
            # init step
            p_res = time_block.copy()
            Phi, autopow = self._delay_and_sum(num, p_res, d_interp2, d_index, amp)
            if 'Cleant' not in self.__class__.__name__:
                if 'Sq' not in self.__class__.__name__:
                    yield Phi[:num]
                elif self.r_diag:
                    yield (Phi[:num] ** 2 - autopow[:num]).clip(min=0)
                else:
                    yield Phi[:num] ** 2
            else:
                # choose correct distance
                blockrm1 = blockrmconv if self.conv_amp else blockrm
                Gamma = zeros(Phi.shape, dtype=fdtype)
                Gamma_autopow = zeros(Phi.shape, dtype=fdtype)
                J = 0
                t_ind = arange(p_res.shape[0], dtype=idtype)
                # deconvolution
                while self.n_iter > J:
                    # print(f"start clean iteration {J+1} of max {self.n_iter}")
                    if self.r_diag:
                        powPhi = (Phi[:num] * Phi[:num] - autopow).sum(0).clip(min=0)
                    else:
                        powPhi = (Phi[:num] * Phi[:num]).sum(0)
                    # find index of max power focus point
                    imax = argmax(powPhi)
                    # find backward delays
                    t_float = (d_interp2[:num, imax, m_index] + d_index[:num, imax, m_index] + n_index).astype(fdtype)
                    # determine max/min delays in sample units
                    # + 2 because we do not want to extrapolate behind the last sample
                    ind_max = t_float.max(0).astype(idtype) + 2
                    ind_min = t_float.min(0).astype(idtype)
                    # store time history at max power focus point
                    h = Phi[:num, imax] * blockr0[:num, imax]
                    for m in range(num_mics):
                        # subtract interpolated time history from microphone signals
                        p_res[ind_min[m] : ind_max[m], m] -= self.damp * interp(
                            t_ind[ind_min[m] : ind_max[m]],
                            t_float[:num, m],
                            h / blockrm1[:num, imax, m],
                        )
                    nextPhi, nextAutopow = self._delay_and_sum(num, p_res, d_interp2, d_index, amp)
                    if self.r_diag:
                        pownextPhi = (nextPhi[:num] * nextPhi[:num] - nextAutopow).sum(0).clip(min=0)
                    else:
                        pownextPhi = (nextPhi[:num] * nextPhi[:num]).sum(0)
                    # print(f"total signal power: {powPhi.sum()}")
                    if pownextPhi.sum() < powPhi.sum():  # stopping criterion
                        Gamma[:num, imax] += self.damp * Phi[:num, imax]
                        Gamma_autopow[:num, imax] = autopow[:num, imax].copy()
                        Phi = nextPhi
                        autopow = nextAutopow
                        # print(f"clean max: {L_p((Gamma**2).sum(0)/num).max()} dB")
                        J += 1
                    else:
                        break
                if 'Sq' not in self.__class__.__name__:
                    yield Gamma[:num]
                elif self.r_diag:
                    yield Gamma[:num] ** 2 - (self.damp**2) * Gamma_autopow[:num]
                else:
                    yield Gamma[:num] ** 2

    def _delay_and_sum(self, num, p_res, d_interp2, d_index, amp):
        """Standard delay-and-sum method."""
        fdtype = float64 if self.precision == 64 else float32
        result = empty((num, self.steer.grid.size), dtype=fdtype)  # output array
        autopow = empty((num, self.steer.grid.size), dtype=fdtype)  # output array
        _delayandsum5(p_res, d_index, d_interp2, amp, result, autopow)
        return result, autopow


class BeamformerTimeSqTraj(BeamformerTimeSq, BeamformerTimeTraj):
    """Provides a time domain beamformer with time-dependent
    power signal output and possible autopower removal
    for a grid moving along a trajectory.
    """

    # internal identifier
    digest = Property(
        depends_on=[
            'steer.digest',
            'source.digest',
            'r_diag',
            'weights',
            'precision',
            'rvec',
            'conv_amp',
            'trajectory.digest',
        ],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=2048):
        """Python generator that yields the **squared** time-domain beamformer output.

        The squared output time signal starts for source signals that were emitted from
        the :class:`~acoular.grids.Grid` at `t=0`.

        Parameters
        ----------
        num : int
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). Defaults to 2048.

        Yields
        ------
        numpy.ndarray
            Samples in blocks of shape (num, :attr:`~BeamformerTime.num_channels`).
                :attr:`~BeamformerTime.num_channels` is usually very \
                large (number of grid points).
                The last block returned by the generator may be shorter than num.
        """
        return super().result(num)


class BeamformerCleant(BeamformerTime):
    """CLEANT deconvolution method.

    An implementation of the CLEAN method in time domain. This class can only
    be used for static sources. See :cite:`Cousson2019` for details.
    """

    #: Boolean flag, always False
    r_diag = Enum(False, desc='False, as we do not remove autopower in this beamformer')

    #: iteration damping factor also referred as loop gain in Cousson et al.
    #: defaults to 0.6
    damp = Range(0.01, 1.0, 0.6, desc='damping factor (loop gain)')

    #: max number of iterations
    n_iter = Int(100, desc='maximum number of iterations')

    # internal identifier
    digest = Property(
        depends_on=['steer.digest', 'source.digest', 'weights', 'damp', 'n_iter'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=2048):
        """Python generator that yields the deconvolved time-domain beamformer output.

        The output starts for signals that were emitted from the :class:`~acoular.grids.Grid` at
        `t=0`.

        Parameters
        ----------
        num : int
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). Defaults to 2048.

        Yields
        ------
        numpy.ndarray
            Samples in blocks of shape (num, :attr:`~BeamformerTime.num_channels`).
                :attr:`~BeamformerTime.num_channels` is usually very \
                large (number of grid points).
                The last block returned by the generator may be shorter than num.
        """
        return super().result(num)


class BeamformerCleantSq(BeamformerCleant):
    """CLEANT deconvolution method with optional removal of autocorrelation.

    An implementation of the CLEAN method in time domain. This class can only
    be used for static sources. See :cite:`Cousson2019` for details on the method
    and :cite:`Kujawski2020` for details on the autocorrelation removal.
    """

    #: Boolean flag, if 'True' (default), the main diagonal is removed before beamforming.
    r_diag = Bool(True, desc='removal of diagonal')

    # internal identifier
    digest = Property(
        depends_on=['steer.digest', 'source.digest', 'weights', 'damp', 'n_iter', 'r_diag'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=2048):
        """Python generator that yields the *squared* deconvolved time-domain beamformer output.

        The output starts for signals that were emitted from the :class:`~acoular.grids.Grid` at
        `t=0`. Per default, block-wise removal of autocorrelation is performed, which can be turned
        of by setting :attr:`r_diag` to `False`.

        Parameters
        ----------
        num : int
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). Defaults to 2048.

        Yields
        ------
        numpy.ndarray
            Samples in blocks of shape (num, :attr:`~BeamformerTime.num_channels`).
                :attr:`~BeamformerTime.num_channels` is usually very \
                large (number of grid points).
                The last block returned by the generator may be shorter than num.
        """
        return super().result(num)


class BeamformerCleantTraj(BeamformerCleant, BeamformerTimeTraj):
    """CLEANT deconvolution method.

    An implementation of the CLEAN method in time domain for moving sources
    with known trajectory. See :cite:`Cousson2019` for details.
    """

    #: Floating point and integer precision
    precision = Enum(32, 64, desc='numeric precision')

    # internal identifier
    digest = Property(
        depends_on=[
            'steer.digest',
            'source.digest',
            'weights',
            'precision',
            'damp',
            'n_iter',
            'rvec',
            'conv_amp',
            'trajectory.digest',
        ],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=2048):
        """Python generator that yields the deconvolved time-domain beamformer output.

        The output starts for signals that were emitted from the :class:`~acoular.grids.Grid` at
        `t=0`.

        Parameters
        ----------
        num : int
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). Defaults to 2048.

        Yields
        ------
        numpy.ndarray
            Samples in blocks of shape (num, :attr:`~BeamformerTime.num_channels`).
                :attr:`~BeamformerTime.num_channels` is usually very \
                large (number of grid points).
                The last block returned by the generator may be shorter than num.
        """
        return super().result(num)


class BeamformerCleantSqTraj(BeamformerCleantTraj, BeamformerTimeSq):
    """CLEANT deconvolution method with optional removal of autocorrelation.

    An implementation of the CLEAN method in time domain for moving sources
    with known trajectory. See :cite:`Cousson2019` for details on the method and
    :cite:`Kujawski2020` for details on the autocorrelation removal.
    """

    #: Boolean flag, if 'True' (default), the main diagonal is removed before beamforming.
    r_diag = Bool(True, desc='removal of diagonal')

    # internal identifier
    digest = Property(
        depends_on=[
            'steer.digest',
            'source.digest',
            'weights',
            'precision',
            'damp',
            'n_iter',
            'rvec',
            'conv_amp',
            'trajectory.digest',
            'r_diag',
        ],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=2048):
        """Python generator that yields the *squared* deconvolved time-domain beamformer output.

        The output starts for signals that were emitted from the :class:`~acoular.grids.Grid` at
        `t=0`. Per default, block-wise removal of autocorrelation is performed, which can be turned
        of by setting :attr:`r_diag` to `False`.

        Parameters
        ----------
        num : int
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). Defaults to 2048.

        Yields
        ------
        numpy.ndarray
            Samples in blocks of shape (num, :attr:`~BeamformerTime.num_channels`).
                :attr:`~BeamformerTime.num_channels` is usually very \
                large (number of grid points).
                The last block returned by the generator may be shorter than num.
        """
        return super().result(num)


class IntegratorSectorTime(TimeOut):
    """Provides an Integrator in the time domain."""

    #: Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    #: :class:`~acoular.grids.RectGrid` object that provides the grid locations.
    grid = Instance(RectGrid, desc='beamforming grid')

    #: List of sectors in grid
    sectors = List()

    #: Clipping, in Decibel relative to maximum (negative values)
    clip = Float(-350.0)

    #: Number of channels in output (= number of sectors).
    num_channels = Property(depends_on=['sectors'])

    # internal identifier
    digest = Property(
        depends_on=['sectors', 'clip', 'grid.digest', 'source.digest'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_num_channels(self):
        return len(self.sectors)

    def result(self, num=1):
        """Python generator that yields the source output integrated over the given
        sectors, block-wise.

        Parameters
        ----------
        num : integer, defaults to 1
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, :attr:`num_channels`).
        :attr:`num_channels` is the number of sectors.
        The last block may be shorter than num.
        """
        inds = [self.grid.indices(*sector) for sector in self.sectors]
        gshape = self.grid.shape
        o = empty((num, self.num_channels), dtype=float)  # output array
        for r in self.source.result(num):
            ns = r.shape[0]
            mapshape = (ns,) + gshape
            rmax = r.max()
            rmin = rmax * 10 ** (self.clip / 10.0)
            r = where(r > rmin, r, 0.0)
            for i, ind in enumerate(inds):
                h = r[:].reshape(mapshape)[(s_[:],) + ind]
                o[:ns, i] = h.reshape(h.shape[0], -1).sum(axis=1)
                i += 1
            yield o[:ns]
